"""The module for Insertion Validation."""
from typing import List, Dict, Optional
import logging

from ga4gh.vrsatile.pydantic.vrs_models import RelativeCopyClass

from variation.schemas.classification_response_schema import Classification
from variation.schemas.app_schemas import Endpoint
from variation.schemas.token_response_schema import Token
from variation.validators.validator import Validator
from variation.schemas.normalize_response_schema\
    import HGVSDupDelMode as HGVSDupDelModeEnum

logger = logging.getLogger("variation")
logger.setLevel(logging.DEBUG)


class InsertionBase(Validator):
    """The Insertion Validator Base class."""

    async def get_valid_invalid_results(
        self, classification_tokens: List, transcripts: List,
        classification: Classification, results: List, gene_tokens: List,
        mane_data_found: Dict, is_identifier: bool,
        hgvs_dup_del_mode: HGVSDupDelModeEnum,
        endpoint_name: Optional[Endpoint] = None,
        baseline_copies: Optional[int] = None,
        relative_copy_class: Optional[RelativeCopyClass] = None,
        do_liftover: bool = False
    ) -> None:
        """Add validation result objects to a list of results.

        :param List classification_tokens: A list of classification Tokens
        :param List transcripts: A list of transcript accessions
        :param Classification classification: A classification for a list of
            tokens
        :param List results: Stores validation result objects
        :param List gene_tokens: List of GeneMatchTokens for a classification
        :param Dict mane_data_found: MANE Transcript information found
        :param bool is_identifier: `True` if identifier is given for exact
            location. `False` otherwise.
        :param HGVSDupDelModeEnum hgvs_dup_del_mode: Must be: `default`, `cnv`,
            `repeated_seq_expr`, `literal_seq_expr`.
            This parameter determines how to represent HGVS dup/del expressions
            as VRS objects.
        :param Optional[Endpoint] endpoint_name: Then name of the endpoint being used
        :param Optional[int] baseline_copies: Baseline copies number
        :param Optional[RelativeCopyClass] relative_copy_class: The relative copy class
        :param bool do_liftover: Whether or not to liftover to GRCh38 assembly
        """
        valid_alleles = list()
        for s in classification_tokens:
            for t in transcripts:
                errors = list()
                t = self.get_accession(t, classification)
                allele = None

                if s.coordinate_type == "c":
                    cds_start_end = await self.uta.get_cds_start_end(t)
                    if cds_start_end is not None:
                        cds_start = cds_start_end[0]
                    else:
                        cds_start = 0
                        allele = None
                        errors.append(f"Unable to get CDS start for {t}")
                else:
                    cds_start = None

                if not errors:
                    allele = self.vrs.to_vrs_allele(
                        t, s.start_pos_flank, s.end_pos_flank,
                        s.coordinate_type, s.alt_type, errors,
                        cds_start=cds_start, alt=s.inserted_sequence
                    )

                if not errors:
                    self.check_pos_index(t, s, errors)

                if not errors and endpoint_name == Endpoint.NORMALIZE:
                    mane = await self.mane_transcript.get_mane_transcript(
                        t, s.start_pos_flank, s.coordinate_type,
                        end_pos=s.end_pos_flank,
                        gene=gene_tokens[0].token if gene_tokens else None,
                        try_longest_compatible=True
                    )
                    self.add_mane_data(mane, mane_data_found, s.coordinate_type,
                                       s.alt_type, s, alt=s.inserted_sequence)

                self.add_validation_result(allele, valid_alleles, results,
                                           classification, s, t, gene_tokens, errors)

                if is_identifier:
                    break

        if endpoint_name == Endpoint.NORMALIZE:
            self.add_mane_to_validation_results(mane_data_found, valid_alleles, results,
                                                classification, gene_tokens)

    def get_hgvs_expr(self, classification: Classification, t: str, s: Token,
                      is_hgvs: bool) -> str:
        """Return a HGVS expression.

        :param Classification classification: A classification for a list of
            tokens
        :param str t: Transcript retrieved from transcript mapping
        :param Token s: The classification token
        :param bool is_hgvs: Whether or not classification is HGVS token
        :return: hgvs expression
        """
        if not is_hgvs:
            prefix = f"{t}:{s.coordinate_type.lower()}."
            position = f"{s.start_pos_flank}_{s.end_pos_flank}"
            if s.inserted_sequence2 is not None:
                inserted_sequence = f"{s.inserted_sequence}_{s.inserted_sequence2}"
            else:
                inserted_sequence = f"{s.inserted_sequence}"

            hgvs_expr = f"{prefix}{position}ins{inserted_sequence}"
        else:
            hgvs_token = [t for t in classification.all_tokens if
                          isinstance(t, Token) and t.token_type
                          in ["HGVS", "ReferenceSequence"]][0]
            hgvs_expr = hgvs_token.input_string
        return hgvs_expr

    def check_pos_index(self, t: str, s: Token, errors: List) -> None:
        """Check that position exists on transcript.

        :param str t: Transcript accession
        :param Token s: Classification token
        :param List errors: List of errors
        """
        sequence, w = self.seqrepo_access.get_reference_sequence(
            t, int(s.start_pos_flank), int(s.end_pos_flank))

        if sequence is None:
            errors.append(w)



"""Module for testing Genomic Substitution Classifier."""
import unittest

from variation.classifiers import GenomicSubstitutionClassifier
from .classifier_base import ClassifierBase


class TestGenomicSubstitutionClassifier(ClassifierBase, unittest.TestCase):
    """A class to test the Coding DNA Substitution Classifier."""

    def classifier_instance(self):
        """Return GenomicSubstitutionClassifier instance."""
        return GenomicSubstitutionClassifier()

    def fixture_name(self):
        """Return GenomicSubstitutionClassifier fixture name."""
        return "genomic_substitution"



"""Module for testing Coding DNA Substitution Translator."""
import unittest

from variation.classifiers import CodingDNASubstitutionClassifier
from variation.translators import CodingDNASubstitution
from variation.validators import CodingDNASubstitution as CDNASUB_V
from .translator_base import TranslatorBase


class TestCodingDNASubstitutionTranslator(TranslatorBase, unittest.TestCase):
    """A class to test the Coding DNA Substitution Translator."""

    def classifier_instance(self):
        """Return coding DNA substitution instance."""
        return CodingDNASubstitutionClassifier()

    def validator_instance(self):
        """Return coding DNA substitution instance."""
        return CDNASUB_V(*self.params)

    def translator_instance(self):
        """Return coding DNA substitution instance."""
        return CodingDNASubstitution()

    def fixture_name(self):
        """Return the fixture name for coding DNA substitution."""
        return "coding_dna_substitution"



"""A module for the Genomic Deletion Classifier."""
from typing import List

from variation.schemas.classification_response_schema import ClassificationType
from .set_based_classifier import SetBasedClassifier


class GenomicDeletionClassifier(SetBasedClassifier):
    """The Genomic Deletion Classifier class."""

    def classification_type(self) -> ClassificationType:
        """Return the Genomic Deletion classification type."""
        return ClassificationType.GENOMIC_DELETION

    def exact_match_candidates(self) -> List[List[str]]:
        """Return the exact match token type candidates."""
        return [
            ["Chromosome", "GenomicDeletion"],
            ["GeneSymbol", "ProteinSubstitution", "GenomicDeletion"],
            ["GenomicDeletion", "GeneSymbol"],
            ["GeneSymbol", "GenomicDeletion"],
            ["HGVS", "GenomicDeletion"],
            ["ReferenceSequence", "GenomicDeletion"],
            ["LocusReferenceGenomic", "GenomicDeletion"]
        ]



"""Module for hgvs_dup_del_mode in normalize endpoint."""
import logging
from typing import Optional, Dict, Tuple, List, Union

from ga4gh.vrs import models
from ga4gh.core import ga4gh_identify
from uta_tools.data_sources import SeqRepoAccess

from variation.schemas.hgvs_to_copy_number_schema import CopyNumberType, \
    RelativeCopyClass
from variation.schemas.normalize_response_schema\
    import HGVSDupDelMode as HGVSDupDelModeEnum

logger = logging.getLogger("variation")
logger.setLevel(logging.DEBUG)


class HGVSDupDelMode:
    """Class for handling how to interpret HGVS duplications and deletions."""

    def __init__(self, seqrepo_access: SeqRepoAccess) -> None:
        """Initialize HGVS Dup Del Mode.

        :param SeqRepoAccess seqrepo_access: Access to seqrepo
        """
        self.seqrepo_access = seqrepo_access
        self.valid_dup_del_modes = {mode.value for mode in
                                    HGVSDupDelModeEnum.__members__.values()}
        self.valid_copy_number_modes = {cn_type.value for cn_type in
                                        CopyNumberType.__members__.values()}

    def is_valid_dup_del_mode(self, mode: str) -> bool:
        """Determine if mode is a valid input for HGVS Dup Del Mode.

        :param str mode: Entered mode
        :return: `True` if valid mode. `False` otherwise.
        """
        hgvs_dup_del_mode = mode.strip().lower()
        return hgvs_dup_del_mode in self.valid_dup_del_modes

    def is_valid_copy_number_mode(self, mode: str) -> bool:
        """Determine if mode is a valid input for copy number mode

        :param str mode: Entered mode
        :return: `True` if valid mode. `False` otherwise.
        """
        copy_number_type_mode = mode.strip().lower()
        return copy_number_type_mode in self.valid_copy_number_modes

    def default_mode(self, ac: str, alt_type: str, pos: Tuple[int, int],
                     del_or_dup: str, location: Dict,
                     chromosome: str = None,
                     allele: Dict = None) -> Optional[Dict]:
        """Use default characteristics to return a variation.
        If endpoints are ambiguous: cnv
            handling X chromosome, make cnv a definite range with base 1-2
            handling Y chromosome, base of 1
            handling anything else, base of 2
        elif len del or dup > 100bp:
            repeated_seq_expr with a derived_seq_expr subject
        else:
            literal_seq_expr (normalized LiteralSequenceExpression Allele)

        :param str ac: Accession
        :param str alt_type: Alteration type
        :param tuple pos: start_pos, end_pos
        :param str del_or_dup: Must be either `del` or `dup`
        :param Dict location: Sequence Location object
        :param str chromosome: Chromosome
        :param Dict allele: VRS Allele object represented as a dict
        :return: VRS Variation object represented as a dict
        """
        if "uncertain" in alt_type or "range" in alt_type:
            variation = self.cnv_mode(ac, del_or_dup,
                                      location, chromosome=chromosome)
        elif pos and (pos[1] - pos[0] > 100):
            variation = self.repeated_seq_expr_mode(alt_type, location)
        else:
            variation = self.literal_seq_expr_mode(allele, alt_type)
        return variation

    def cnv_mode(self, ac: str, del_or_dup: str, location: Dict,
                 chromosome: str = None) -> Optional[Dict]:
        """Return a VRS Copy Number Variation.

        :param str ac: Accession
        :param str del_or_dup: Must be either `del` or `dup`
        :param Dict location: VRS SequenceLocation
        :param str chromosome: Chromosome
        :return: VRS Copy Number object represented as a dict
        """
        if chromosome is None:
            chromosome, _ = self.seqrepo_access.ac_to_chromosome(ac)

        if chromosome is None:
            logger.warning(f"Unable to find chromosome on {ac}")
            return None
        if chromosome == "X":
            copies = models.DefiniteRange(
                min=0 if del_or_dup == "del" else 2,
                max=1 if del_or_dup == "del" else 3,
                type="DefiniteRange"
            )
        elif chromosome == "Y":
            copies = models.Number(
                value=0 if del_or_dup == "del" else 2, type="Number"
            )
        else:
            # Chr 1-22
            copies = models.Number(
                value=1 if del_or_dup == "del" else 3, type="Number"
            )
        self._abs_cnv(location, copies)

        variation = models.AbsoluteCopyNumber(
            subject=models.DerivedSequenceExpression(
                location=location,
                reverse_complement=False,
                type="DerivedSequenceExpression"
            ),
            copies=copies,
            type="AbsoluteCopyNumber"
        )
        return self._ga4gh_identify_variation(variation)

    def repeated_seq_expr_mode(self, alt_type: str,
                               location: Dict) -> Optional[Dict]:
        """Return a VRS Allele with a RepeatedSequenceExpression.
        The RepeatedSequenceExpression subject will be a
            DerivedSequenceExpression.

        :param str alt_type: Alteration type
        :param Dict location: VRS SequenceLocation
        :return: VRS Allele object represented as a dict
        """
        if "range" in alt_type:
            # Ranges should return an error
            return None

        if alt_type == "duplication":
            count = models.Number(value=2, type="Number")
        elif alt_type == "deletion":
            count = models.Number(value=0, type="Number")
        else:
            return None

        seq_expr = models.RepeatedSequenceExpression(
            seq_expr=models.DerivedSequenceExpression(
                location=location,
                reverse_complement=False,
                type="DerivedSequenceExpression"
            ),
            count=count,
            type="RepeatedSequenceExpression"
        )

        variation = models.Allele(
            location=location,
            state=seq_expr,
            type="Allele"
        )
        return self._ga4gh_identify_variation(variation)

    def literal_seq_expr_mode(self, allele: Dict,
                              alt_type: str) -> Optional[Dict]:
        """Return a VRS Allele with a normalized LiteralSequenceExpression.

        :param Dict allele: normalized VRS Allele object represented as a dict
        :param str alt_type: Alteration type
        :return: VRS Allele object represented as a dict
        """
        if "range" in alt_type or "uncertain" in alt_type:
            return None

        variation = models.Allele(**allele) if allele else None
        return self._ga4gh_identify_variation(variation)

    @staticmethod
    def _ga4gh_identify_variation(variation: models.Variation) -> Optional[Dict]:
        """Return variation with GA4GH digest-based id.

        :param models.Variation variation: VRS variation object
        :return: VRS Variation with GA4GH digest-based id represented as a dict
        """
        if variation is None:
            return None
        else:
            variation._id = ga4gh_identify(variation)
            return variation.as_dict()

    def interpret_variation(
        self, ac: str, alt_type: str, allele: Dict, errors: List,
        hgvs_dup_del_mode: HGVSDupDelModeEnum, pos: Optional[Tuple[int, int]] = None,
        baseline_copies: Optional[int] = None,
        relative_copy_class: Optional[RelativeCopyClass] = None
    ) -> Dict:
        """Interpret variation using HGVSDupDelMode

        :param str ac: Accession
        :param str alt_type: Alteration type
        :param Dict allele: VRS Allele object
        :param List errors: List of errors
        :param HGVSDupDelModeEnum hgvs_dup_del_mode: Mode to use for
            interpreting HGVS duplications and deletions
        :param Optional[Tuple[int, int]] pos: Position changes
        :param Optional[int] baseline_copies: Baseline copies number
        :param Optional[RelativeCopyClass] relative_copy_class: The relative copy class
        :return: VRS Variation object
        """
        if "deletion" in alt_type:
            del_or_dup = "del"
        else:
            del_or_dup = "dup"
        variation = None
        if allele is None:
            errors.append("Unable to get Allele")
        else:
            if hgvs_dup_del_mode == HGVSDupDelModeEnum.DEFAULT:
                variation = self.default_mode(
                    ac, alt_type, pos, del_or_dup,
                    allele["location"], allele=allele
                )
            elif hgvs_dup_del_mode == HGVSDupDelModeEnum.CNV:
                variation = self.cnv_mode(ac, del_or_dup, allele["location"])
            elif hgvs_dup_del_mode == HGVSDupDelModeEnum.REPEATED_SEQ_EXPR:
                variation = self.repeated_seq_expr_mode(
                    alt_type, allele["location"]
                )
            elif hgvs_dup_del_mode == HGVSDupDelModeEnum.LITERAL_SEQ_EXPR:
                variation = self.literal_seq_expr_mode(allele, alt_type)
            elif hgvs_dup_del_mode == CopyNumberType.ABSOLUTE:
                variation = self.absolute_copy_number_mode(
                    ac, del_or_dup, allele["location"], baseline_copies=baseline_copies)
            elif hgvs_dup_del_mode == CopyNumberType.RELATIVE:
                variation = self.relative_copy_number_mode(
                    allele["location"], relative_copy_class)
            if not variation:
                errors.append("Unable to get VRS Variation")
        return variation

    def _abs_cnv(
        self, location: Dict,
        copies: Union[models.DefiniteRange, models.Number, models.IndefiniteRange]
    ) -> Dict:
        """Return absolute copy number variation with ga4gh digest as optional id

        :param Union[models.DefiniteRange, models.Number, models.IndefiniteRange] copies:  # noqa E501
            Copies for absolute copy number variation
        :param Dict location: VRS SequenceLocation
        :return: VRS Absolute Copy Number object represented as a dict
        """
        variation = models.AbsoluteCopyNumber(
            subject=models.DerivedSequenceExpression(
                location=location,
                reverse_complement=False,
                type="DerivedSequenceExpression"
            ),
            copies=copies,
            type="AbsoluteCopyNumber"
        )
        return self._ga4gh_identify_variation(variation)

    def absolute_copy_number_mode(
        self, ac: str, del_or_dup: str, location: Dict,
        chromosome: Optional[str] = None, baseline_copies: Optional[int] = None
    ) -> Optional[Dict]:
        """Return absolute copy number variation

        :param str ac: Accession
        :param str del_or_dup: Must be either `del` or `dup`
        :param Dict location: VRS SequenceLocation
        :param str chromosome: Chromosome
        :param Optional[int] baseline_copies: Baseline copies number
        :return: Absolute copy number variation as a dict
        """
        if baseline_copies:
            copies = models.Number(
                value=baseline_copies - 1 if del_or_dup == "del" else baseline_copies + 1,  # noqa: E501
                type="Number"
            )
            return self._abs_cnv(location, copies)
        else:
            return self.cnv_mode(ac, del_or_dup, location, chromosome)

    def relative_copy_number_mode(
        self, location: Dict, relative_copy_class: RelativeCopyClass
    ) -> Optional[Dict]:
        """Return relative copy number variation

        :param Dict location: VRS SequenceLocation
        :param RelativeCopyClass relative_copy_class: The relative copy class
        :return: Relative copy number variation as a dict
        """
        variation = models.RelativeCopyNumber(
            type="RelativeCopyNumber",
            subject=models.DerivedSequenceExpression(
                location=location, reverse_complement=False,
                type="DerivedSequenceExpression"
            ),
            relative_copy_class=relative_copy_class
        )
        return self._ga4gh_identify_variation(variation)



"""A module for Insertion Tokenization Base Class."""
from abc import abstractmethod
from typing import Optional, Dict, List

from variation.schemas.token_response_schema import Insertion, TokenMatchType, Token
from .tokenizer import Tokenizer
from .caches import AminoAcidCache, NucleotideCache
from .tokenize_base import TokenizeBase


class InsertionBase(Tokenizer):
    """Class for tokenizing Insertion."""

    def __init__(self, amino_acid_cache: AminoAcidCache,
                 nucleotide_cache: NucleotideCache) -> None:
        """Initialize the Insertion Class.

        :param AminoAcidCache amino_acid_cache: Valid amino acid codes
        :param NucleotideCache nucleotide_cache: Valid nucleotides
        """
        self.parts = None
        self.tokenize_base = TokenizeBase(amino_acid_cache, nucleotide_cache)

    def match(self, input_string: str) -> Optional[Insertion]:
        """Return tokens that match the input string."""
        if input_string is None:
            return None

        self.parts = {
            "token": input_string,
            "input_string": input_string,
            "match_type": TokenMatchType.UNSPECIFIED.value,
            "start_pos_flank": None,
            "end_pos_flank": None,
            "inserted_sequence": None,
            "inserted_sequence2": None,
            "coordinate_type": None
        }

        input_string = str(input_string).lower()

        if input_string.startswith("(") and input_string.endswith(")"):
            input_string = input_string[1:-1]

        conditions = (
            "ins" in input_string,
            "del" not in input_string and "delins" not in input_string,
            input_string.startswith("c.") or input_string.startswith("g.")
        )
        if not all(conditions):
            return None

        parts = input_string.split("ins")
        self._get_parts(parts)
        return self.return_token(self.parts)

    def _get_parts(self, parts: List) -> None:
        """Get parts for Insertion.

        :param List parts: Parts of input string
        """
        if len(parts) != 2:
            return None

        if not parts[0].startswith("c.") and not parts[0].startswith("g."):
            return None

        # Get reference sequence
        coordinate_type = parts[0][:1]
        parts[0] = parts[0][2:]

        positions = self.tokenize_base.get_positions_deleted(parts)
        if not positions:
            return None
        start_pos = positions[0]
        end_pos = positions[1]

        inserted_sequences = \
            self.tokenize_base.get_transcript_genomic_inserted_sequence(parts)
        if not inserted_sequences or not inserted_sequences[0]:
            return None

        if "_" in parts[0] and parts[0].count("_") ==\
                2 and not inserted_sequences[1]:
            return None

        self.parts["start_pos_flank"] = start_pos
        self.parts["end_pos_flank"] = end_pos
        self.parts["inserted_sequence"] = inserted_sequences[0]
        self.parts["inserted_sequence2"] = inserted_sequences[1]
        self.parts["coordinate_type"] = coordinate_type

    @abstractmethod
    def return_token(self, params: Dict[str, str]) -> Optional[Token]:
        """Return token instance."""
        raise NotImplementedError



"""Module for testing Genomic Deletion Range Classifier."""
import unittest

from variation.classifiers import GenomicDeletionRangeClassifier
from .classifier_base import ClassifierBase


class TestGenomicDeletionRangeClassifier(ClassifierBase,
                                         unittest.TestCase):
    """A class to test the Genomic Deletion Range Classifier."""

    def classifier_instance(self):
        """Return Genomic Deletion Range Classifier instance."""
        return GenomicDeletionRangeClassifier()

    def fixture_name(self):
        """Return Genomic Deletion Range fixture name."""
        return "genomic_deletion_range"



"""Module for commonly used tokenization methods."""
from typing import Tuple, Optional, Union, List
import re

from variation.tokenizers.caches import NucleotideCache, AminoAcidCache


class TokenizeBase:
    """Class for Tokenize methods."""

    def __init__(self, amino_acid_cache: AminoAcidCache,
                 nucleotide_cache: NucleotideCache) -> None:
        """Initialize Token Base class.

        :param AminoAcidCache amino_acid_cache: Valid amino acid codes
        :param NucleotideCache nucleotide_cache: Valid nucleotides
        """
        self.nucleotide_cache = nucleotide_cache
        self.amino_acid_cache = amino_acid_cache
        self.splitter_char_digit = re.compile("([a-zA-Z]+)([0-9]+)")

    def get_amino_acid_and_pos(
        self, part: List, used_one_letter: bool
    ) -> Optional[Tuple[str, int, bool]]:
        """Return amino acid and position.

        :param List part: Tokenized input string
        :param bool used_one_letter: `True` if used 1 letter AA code.
            `False` if used 3 letter AA code.
        :return: Three letter AA code, position, and whether or not
            one letter AA code was used
        """
        try:
            char_and_digits = self.splitter_char_digit.match(part).groups()
        except AttributeError:
            return None

        if len(char_and_digits) != 2 or len(part) != \
                (len(char_and_digits[0]) + len(char_and_digits[1])):
            return None

        aa = char_and_digits[0]
        if len(aa) == 1:
            if not used_one_letter:
                used_one_letter = True
            tmp_aa = \
                self.amino_acid_cache.amino_acid_code_conversion[aa.upper()]
        else:
            tmp_aa = aa
        pos = char_and_digits[1]

        if not self.amino_acid_cache.__contains__(tmp_aa) \
                or not pos.isdigit():
            return None
        return aa.upper(), pos, used_one_letter

    def get_protein_inserted_sequence(self, parts: List,
                                      used_one_letter: bool) -> Optional[str]:
        """Return inserted sequence for protein reference sequence.

        :param List parts: Tokenized input string
        :param bool used_one_letter: `True` if used 1 letter AA code.
            `False` if used 3 letter AA code.
        :return: Inserted sequence
        """
        # Check inserted sequences
        inserted_sequence = ""
        if used_one_letter:
            for i in range(len(parts[1])):
                aa = parts[1][i:i + 1]
                if len(aa) != 1:
                    return None
                try:
                    self.amino_acid_cache.amino_acid_code_conversion[aa.upper()]  # noqa: E501
                except KeyError:
                    return None
                else:
                    inserted_sequence += aa.upper()
        else:
            for i in range(0, len(parts[1]), 3):
                aa = parts[1][i:i + 3]
                if len(aa) != 3 or not self.amino_acid_cache.__contains__(aa):
                    if aa != "ter":
                        return None
                inserted_sequence += \
                    self.amino_acid_cache.convert_three_to_one(aa)

        if inserted_sequence == "":
            return None
        return inserted_sequence

    def get_aa_pos_range(self,
                         parts: List) -> Optional[Tuple[str, str, str, int, bool]]:
        """Get amino acid(s) and positions(s) for protein reference sequence.

        :param List parts: Tokenized input string
        :return: Beginning AA, End AA,  Beginning position, End position,
            Whether or not one letter code was used
        """
        aa_start = None
        aa_end = None
        pos_start = None
        pos_end = None
        used_one_letter = False

        if "_" in parts[0] and parts[0].count("_") == 1:
            aa_pos_range = parts[0].split("_")
            if len(aa_pos_range) != 2 or \
                    not aa_pos_range[0] or not aa_pos_range[1]:
                return None

            start_aa_pos = \
                self.get_amino_acid_and_pos(
                    aa_pos_range[0], used_one_letter
                )

            if start_aa_pos:
                used_one_letter = start_aa_pos[2]

            end_aa_pos = \
                self.get_amino_acid_and_pos(
                    aa_pos_range[1], used_one_letter
                )

            if start_aa_pos and end_aa_pos:
                aa_start = start_aa_pos[0]
                pos_start = start_aa_pos[1]
                aa_end = end_aa_pos[0]
                pos_end = end_aa_pos[1]
                used_one_letter = end_aa_pos[2]

        else:
            aa_and_pos = \
                self.get_amino_acid_and_pos(
                    parts[0], used_one_letter
                )
            if aa_and_pos:
                aa_start = aa_and_pos[0]
                pos_start = aa_and_pos[1]
                used_one_letter = aa_and_pos[2]

        return aa_start, aa_end, pos_start, pos_end, used_one_letter

    def get_positions_deleted(self, parts: List) -> Optional[Tuple[str, str]]:
        """Return position(s) deleted for transcript and genomic references.

        :param List parts: Tokenized input string
        :return: Start position deleted and end position deleted
        """
        if "_" in parts[0] and parts[0].count("_") == 1:
            positions = self.get_valid_digits(parts[0])
            if not positions:
                return None
            start_pos_del, end_pos_del = positions
            if start_pos_del > end_pos_del:
                return None
        else:
            start_pos_del = parts[0]
            end_pos_del = None
            if not start_pos_del.isdigit():
                return None
        return start_pos_del, end_pos_del

    def get_transcript_genomic_inserted_sequence(
        self, parts: List
    ) -> Optional[Tuple[Union[str, int], Union[str, int]]]:
        """Return inserted sequence for transcript and genomic references.

        :param List parts: Tokenized input string
        :return: Start inserted sequence and end inserted sequence
        """
        # Check inserted sequences
        if "_" in parts[1] and parts[1].count("_") == 1:
            # Replaced by sequence positions
            inserted_sequences = self.get_valid_digits(parts[1])
            if not inserted_sequences:
                return None
            inserted_sequence1, inserted_sequence2 = inserted_sequences
            if inserted_sequence1 > inserted_sequence2:
                return None
        else:
            # Replaced by nucleotides
            inserted_sequence1 = self.get_sequence(parts[1])
            inserted_sequence2 = None
        return inserted_sequence1, inserted_sequence2

    def get_sequence(self, part: str) -> Optional[str]:
        """Return validated sequence for transcript and genomic references.

        :param str part: Sequence to validate
        :return: Sequence of nucleotides
        """
        for char in part:
            if char.upper() not in self.nucleotide_cache.base_nucleotides:
                return None
        return part.upper()

    def get_valid_digits(self, part: str) -> Optional[Tuple[str, str]]:
        """Return valid digits after splitting on `_`.

        :param str part: Range of digits
        :return: Digits represented as strings
        """
        digits = part.split("_")
        digit1 = digits[0]
        digit2 = digits[1]
        if not digit1.isdigit() or not digit2.isdigit():
            return None
        return digit1, digit2



"""A module for Genomic Duplication Tokenization."""
from typing import Optional, Union, List

from variation.schemas.token_response_schema import DuplicationAltType, \
    GenomicDuplicationToken, GenomicDuplicationRangeToken
from variation.tokenizers.duplication_base import DuplicationBase


class GenomicDuplication(DuplicationBase):
    """Class for tokenizing duplications on the genomic coordinate."""

    def _get_parts(self, parts: List) -> None:
        if len(parts) != 1 or not parts[0].startswith("g."):
            return None

        parts[0] = parts[0][2:]
        if "_" in parts[0]:
            if parts[0].count("_") == 1:
                pos = parts[0].split("_")
                try:
                    pos[0] = int(pos[0])
                    pos[1] = int(pos[1])
                except ValueError:
                    pass
                else:
                    if pos[0] < pos[1]:
                        self.parts["start_pos1_dup"] = pos[0]
                        self.parts["start_pos2_dup"] = pos[1]
                        self.parts["coordinate_type"] = "g"
            else:
                self.parts["alt_type"] = DuplicationAltType.DUPLICATION_RANGE
                parts = parts[0].split("_")
                len_parts = len(parts)
                if len_parts == 4:
                    for part_ix, parts_field in [
                        (0, "start_pos1_dup"),
                        (1, "start_pos2_dup"),
                        (2, "end_pos1_dup"),
                        (3, "end_pos2_dup")
                    ]:
                        part_val = self._check_uncertain_or_int(parts[part_ix])
                        if part_val is None:
                            return None
                        else:
                            self.parts[parts_field] = part_val
                    self.parts["coordinate_type"] = "g"

                elif len_parts == 3:
                    if "(" in parts[0] and ")" in parts[1]:
                        # Format is: (?_#)_#dup
                        for part_ix, parts_field in [
                            (0, "start_pos1_dup"),
                            (1, "start_pos2_dup"),
                            (2, "end_pos1_dup")
                        ]:
                            part_val = self._check_uncertain_or_int(
                                parts[part_ix])
                            if part_val is None:
                                return None
                            else:
                                self.parts[parts_field] = part_val
                    else:
                        # Format is #_(#_?)dup
                        for part_ix, parts_field in [
                            (0, "start_pos1_dup"),
                            (1, "end_pos1_dup"),
                            (2, "end_pos2_dup")
                        ]:
                            part_val = self._check_uncertain_or_int(
                                parts[part_ix])
                            if part_val is None:
                                return None
                            else:
                                self.parts[parts_field] = part_val
                    self.parts["coordinate_type"] = "g"
        else:
            try:
                pos = int(parts[0])
            except ValueError:
                pass
            else:
                self.parts["start_pos1_dup"] = pos
                self.parts["coordinate_type"] = "g"
        return None

    def _check_uncertain_or_int(self, part: str) -> Optional[Union[int, str]]:
        part = part.replace("(", "")
        part = part.replace(")", "")
        try:
            return int(part)
        except ValueError:
            if part == "?":
                self.parts["alt_type"] = \
                    DuplicationAltType.UNCERTAIN_DUPLICATION
                return part
        return None

    def return_token(self) -> Optional[Union[GenomicDuplicationRangeToken,
                                             GenomicDuplicationToken]]:
        """Return token instance if a match is found."""
        # we only set this field if it"s valid
        if self.parts["coordinate_type"] == "g":
            if self.parts["end_pos1_dup"] is None and \
                    self.parts["end_pos2_dup"] is None:

                if self.parts["start_pos2_dup"] and self.parts["start_pos1_dup"] > self.parts["start_pos2_dup"]:  # noqa: E501
                    return

                self.parts["alt_type"] = DuplicationAltType.DUPLICATION
                return GenomicDuplicationToken(**self.parts)
            else:
                prev_val = None
                for field in ["start_pos1_dup", "start_pos2_dup",
                              "end_pos1_dup", "end_pos2_dup"]:
                    val = self.parts[field]
                    if val not in ["?", None]:
                        if prev_val is not None:
                            if prev_val > val:
                                return
                        prev_val = val
                return GenomicDuplicationRangeToken(**self.parts)



"""Module for the set based classifier."""
from typing import List, Optional, Set

from variation.schemas.classification_response_schema import Classification, \
    ClassificationType, ConfidenceRating
from variation.schemas.token_response_schema import Token
from .classifier import Classifier


class SetBasedClassifier(Classifier):
    """The set based classifier class for finding classification matches."""

    def match(self, tokens: List[Token]) -> Optional[Classification]:
        """Return a classification match for a list of tokens.

        :param List[Token] tokens: List of tokens to determine classification
            match type
        :return: A classification for a list of tokens
        """
        token_types = list(map(lambda t: t.token_type, tokens))
        exact_matches: List[List[str]] = []
        out_of_order_matches: List[List[str]] = []
        matches_with_extra_terms: Set[str] = set()
        partial_matches: Set[str] = set()

        for candidate in self.exact_match_candidates():
            token_set = set(token_types)
            candidate_set = set(candidate)
            if token_types == candidate:
                exact_matches.append(candidate)
            elif token_set == candidate_set:
                out_of_order_matches.append(candidate)
            elif token_set > candidate_set:
                matches_with_extra_terms.update(token_set & candidate_set)
            elif len(token_set & candidate_set) > 0:
                partial_matches.update(token_set & candidate_set)

        if len(exact_matches) == 1:
            return Classification(
                classification_type=self.classification_type(),
                matching_tokens=exact_matches[0],
                non_matching_tokens=[],
                all_tokens=tokens,
                confidence=ConfidenceRating.EXACT
            )
        elif len(out_of_order_matches) > 0:
            all_unordered_matches: Set[str] = set()
            for match in out_of_order_matches:
                for token in match:
                    all_unordered_matches.add(token)
            return Classification(
                classification_type=self.classification_type(),
                matching_tokens=list(all_unordered_matches),
                non_matching_tokens=[],
                all_tokens=tokens,
                confidence=ConfidenceRating.UNORDERED
            )
        elif len(matches_with_extra_terms) > 0:
            return Classification(
                classification_type=self.classification_type(),
                matching_tokens=list(matches_with_extra_terms),
                non_matching_tokens=list(token_set - matches_with_extra_terms),
                all_tokens=tokens,
                confidence=ConfidenceRating.SUPERSET
            )
        elif len(partial_matches) > 0:
            return Classification(
                classification_type=self.classification_type(),
                matching_tokens=list(partial_matches),
                non_matching_tokens=list(token_set - partial_matches),
                all_tokens=tokens,
                confidence=ConfidenceRating.INTERSECTION
            )
        else:
            return None

    def classification_type(self) -> ClassificationType:
        """Return the classification type."""
        pass

    def exact_match_candidates(self) -> List[List[str]]:
        """Return the token match candidates for a given classification.

        :return: List of tokens, where order matters, that make up a given
            classification.
        """
        pass



"""A module for the Genomic Insertion Classifier."""
from typing import List

from variation.schemas.classification_response_schema import ClassificationType
from .set_based_classifier import SetBasedClassifier


class GenomicInsertionClassifier(SetBasedClassifier):
    """The Genomic Insertion Classifier class."""

    def classification_type(self) -> ClassificationType:
        """Return the Genomic Insertion classification type."""
        return ClassificationType.GENOMIC_INSERTION

    def exact_match_candidates(self) -> List[List[str]]:
        """Return the exact match token type candidates."""
        return [
            ["Chromosome", "GenomicInsertion"],
            ["GeneSymbol", "ProteinSubstitution", "GenomicInsertion"],
            ["GenomicInsertion", "GeneSymbol"],
            ["GeneSymbol", "GenomicInsertion"],
            ["HGVS", "GenomicInsertion"],
            ["ReferenceSequence", "GenomicInsertion"],
            ["LocusReferenceGenomic", "GenomicInsertion"]
        ]



"""Module for Coding DNA DelIns Translation."""
from variation.translators.translator import Translator
from variation.schemas.classification_response_schema import ClassificationType
from variation.schemas.token_response_schema import CodingDNADelInsToken, Token


class CodingDNADelIns(Translator):
    """The Coding DNA DelIns Translator class."""

    def can_translate(self, type: ClassificationType) -> bool:
        """Return if classification type is Coding DNA DelIns."""
        return type == ClassificationType.CODING_DNA_DELINS

    def is_token_instance(self, token: Token) -> bool:
        """Return if the token is an Coding DNA DelIns token instance."""
        return isinstance(token, CodingDNADelInsToken)



"""Module for Silent Mutation Translation."""
from variation.translators.translator import Translator
from variation.schemas.classification_response_schema import ClassificationType
from variation.schemas.token_response_schema import SilentMutationToken, Token


class SilentMutation(Translator):
    """The Silent Mutation Translator class."""

    def can_translate(self, type: ClassificationType) -> bool:
        """Return if classification type is Silent Mutation."""
        return type == ClassificationType.SILENT_MUTATION

    def is_token_instance(self, token: Token) -> bool:
        """Return if the token is an Silent Mutation token instance."""
        return isinstance(token, SilentMutationToken)



"""Module for translation."""
from typing import List, Optional, Dict

from variation.schemas.validation_response_schema import ValidationResult
from .translator import Translator
from .protein_substitution import ProteinSubstitution
from .polypeptide_truncation import PolypeptideTruncation
from .silent_mutation import SilentMutation
from .coding_dna_substitution import CodingDNASubstitution
from .genomic_substitution import GenomicSubstitution
from .coding_dna_silent_mutation import CodingDNASilentMutation
from .genomic_silent_mutation import GenomicSilentMutation
from .protein_delins import ProteinDelIns
from .coding_dna_delins import CodingDNADelIns
from .genomic_delins import GenomicDelIns
from .protein_deletion import ProteinDeletion
from .coding_dna_deletion import CodingDNADeletion
from .genomic_deletion import GenomicDeletion
from .protein_insertion import ProteinInsertion
from .coding_dna_insertion import CodingDNAInsertion
from .genomic_insertion import GenomicInsertion
from .genomic_uncertain_deletion import GenomicUncertainDeletion
from .genomic_duplication import GenomicDuplication
from .genomic_deletion_range import GenomicDeletionRange


class Translate:
    """The translation class."""

    def __init__(self) -> None:
        """Initialize the translation class."""
        self.all_translators: List[Translator] = [
            ProteinSubstitution(),
            PolypeptideTruncation(),
            SilentMutation(),
            CodingDNASubstitution(),
            GenomicSubstitution(),
            CodingDNASilentMutation(),
            GenomicSilentMutation(),
            ProteinDelIns(),
            CodingDNADelIns(),
            GenomicDelIns(),
            ProteinDeletion(),
            CodingDNADeletion(),
            GenomicDeletion(),
            ProteinInsertion(),
            CodingDNAInsertion(),
            GenomicInsertion(),
            GenomicDeletionRange(),
            GenomicUncertainDeletion(),
            GenomicDuplication()
        ]

    def perform(self, res: ValidationResult) -> Optional[Dict]:
        """Translate a valid variation query."""
        for translator in self.all_translators:
            if translator.can_translate(
                    res.classification.classification_type):
                return translator.translate(res)
        return None



"""A module for testing Genomic Duplication Range Tokenization."""
import unittest

from variation.tokenizers import GenomicDuplication
from .tokenizer_base import TokenizerBase


class TestGenomicDuplicationRangeTokenizer(TokenizerBase, unittest.TestCase):
    """A class for testing Genomic Duplication Tokenization."""

    def tokenizer_instance(self):
        """Return Genomic Duplication instance."""
        return GenomicDuplication()

    def token_type(self):
        """Return genomic duplication token type."""
        return "GenomicDuplicationRange"

    def fixture_name(self):
        """Return the fixture name for Genomic Duplication."""
        return "genomic_duplication_range"



"""A module for testing LRG tokenization."""
import unittest

from variation.tokenizers import LocusReferenceGenomic
from .tokenizer_base import TokenizerBase


class TestLocusReferenceGenomicTokenizer(TokenizerBase, unittest.TestCase):
    """A class for testing LRG Tokenization."""

    def tokenizer_instance(self):
        """Return LRG instance."""
        return LocusReferenceGenomic()

    def token_type(self):
        """Return LRG token type."""
        return "LocusReferenceGenomic"

    def fixture_name(self):
        """Return fixture name for LRG."""
        return "locus_reference_genomic"



"""Module for testing Protein DelIns Validator."""
import unittest

from variation.validators import ProteinDelIns
from variation.classifiers import ProteinDelInsClassifier
from .validator_base import ValidatorBase


class TestProteinDelInsValidator(ValidatorBase, unittest.TestCase):
    """A class to test the Protein DelIns Validator."""

    def validator_instance(self):
        """Return protein delins instance."""
        return ProteinDelIns(*self.aa_params)

    def classifier_instance(self):
        """Return the protein delins classifier instance."""
        return ProteinDelInsClassifier()

    def fixture_name(self):
        """Return the fixture name for protein delins."""
        return "protein_delins"



"""Module for testing Genomic Uncertain Deletion Validator."""
import unittest

from variation.validators import GenomicUncertainDeletion
from variation.classifiers import GenomicUncertainDeletionClassifier
from .validator_base import ValidatorBase


class TestGenomicUncertainDeletionValidator(ValidatorBase, unittest.TestCase):
    """A class to test the Genomic Uncertain Deletion Validator."""

    def validator_instance(self):
        """Return genomic uncertain deletion instance."""
        return GenomicUncertainDeletion(*self.params)

    def classifier_instance(self):
        """Return the genomic uncertain deletion classifier instance."""
        return GenomicUncertainDeletionClassifier()

    def fixture_name(self):
        """Return the fixture name for genomic uncertain deletion."""
        return "genomic_uncertain_deletion"



"""A module for testing Genomic Uncertain Deletion tokenization."""
import unittest

from variation.tokenizers import GenomicUncertainDeletion
from .tokenizer_base import TokenizerBase


class TestGenomicUncertainDeletionTokenizer(TokenizerBase, unittest.TestCase):
    """A class for testing Genomic Uncertain Deletion Tokenization."""

    def tokenizer_instance(self):
        """Return Genomic Uncertain Deletion instance."""
        return GenomicUncertainDeletion()

    def token_type(self):
        """Return genomic uncertain deletion token type."""
        return "GenomicUncertainDeletion"

    def fixture_name(self):
        """Return the fixture name for genomic uncertain deletion."""
        return "genomic_uncertain_deletion"



"""Module for Validation Response Schema."""
from typing import List, Optional
from enum import IntEnum

from pydantic import BaseModel
from pydantic.types import StrictBool

from variation.schemas.classification_response_schema import Classification
from variation.schemas.token_response_schema import GeneMatchToken, Token


class LookupType(IntEnum):
    """IntEnum for Lookup Type."""

    GENE_SYMBOL = 1


class ValidationResult(BaseModel):
    """Validation Results for a given variation."""

    classification: Classification
    classification_token: Optional[Token]
    is_valid: StrictBool
    confidence_score: float
    variation: Optional[dict] = None
    errors: List[str]
    gene_tokens: Optional[List[GeneMatchToken]]
    is_mane_transcript: Optional[StrictBool]
    identifier: Optional[str]


class ValidationSummary(BaseModel):
    """Give Valid and Invalid Results for a given variation."""

    valid_results: List[ValidationResult]
    invalid_results: List[ValidationResult]
    warnings: Optional[List[str]]


class ValidationResponseSchema(BaseModel):
    """Validation Response for a given variation."""

    search_term: str
    validation_summary: ValidationSummary



"""A module for testing Protein DelIns tokenization."""
import unittest

from variation.tokenizers.caches import AminoAcidCache, NucleotideCache
from variation.tokenizers import ProteinDelIns
from .tokenizer_base import TokenizerBase


class TestProteinDelInsTokenizer(TokenizerBase, unittest.TestCase):
    """A class for testing Protein DelIns Tokenization."""

    def tokenizer_instance(self):
        """Return Protein DelIns instance."""
        return ProteinDelIns(AminoAcidCache(), NucleotideCache())

    def token_type(self):
        """Return protein delins token type."""
        return "ProteinDelIns"

    def fixture_name(self):
        """Return the fixture name for protein delins."""
        return "protein_delins"



"""Module for testing Genomic Substitution Validator."""
import unittest

from variation.validators import GenomicSubstitution
from variation.classifiers import GenomicSubstitutionClassifier
from .validator_base import ValidatorBase


class TestGenomicSubstitutionValidator(ValidatorBase, unittest.TestCase):
    """A class to test the Genomic Substitution Validator."""

    def validator_instance(self):
        """Return genomic substitution instance."""
        return GenomicSubstitution(*self.params)

    def classifier_instance(self):
        """Return the genomic substitution classifier instance."""
        return GenomicSubstitutionClassifier()

    def fixture_name(self):
        """Return the fixture name for genomic substitution."""
        return "genomic_substitution"



"""A module for the Coding DNA insertion Classifier."""
from typing import List

from variation.schemas.classification_response_schema import ClassificationType
from .set_based_classifier import SetBasedClassifier


class CodingDNAInsertionClassifier(SetBasedClassifier):
    """The Coding DNA insertion Classifier class."""

    def classification_type(self) -> ClassificationType:
        """Return the Coding DNA insertion classification type."""
        return ClassificationType.CODING_DNA_INSERTION

    def exact_match_candidates(self) -> List[List[str]]:
        """Return the exact match token type candidates."""
        return [
            ["CodingDNAInsertion"],
            ["GeneSymbol", "ProteinSubstitution", "CodingDNAInsertion"],
            ["CodingDNAInsertion", "GeneSymbol"],
            ["GeneSymbol", "CodingDNAInsertion"],
            ["HGVS", "CodingDNAInsertion"],
            ["ReferenceSequence", "CodingDNAInsertion"],
            ["LocusReferenceGenomic", "CodingDNAInsertion"]
        ]



"""Module for toVRS endpoint."""
from typing import Tuple, Optional, List, Union
from urllib.parse import unquote

from gene.query import QueryHandler as GeneQueryHandler
from ga4gh.vrsatile.pydantic.vrs_models import Allele, Haplotype, AbsoluteCopyNumber,\
    VariationSet, Text
from uta_tools.data_sources import SeqRepoAccess, TranscriptMappings, UTADatabase, \
    MANETranscriptMappings, MANETranscript

from variation.hgvs_dup_del_mode import HGVSDupDelMode
from variation.schemas.app_schemas import Endpoint
from variation.schemas.hgvs_to_copy_number_schema import VALID_CLASSIFICATION_TYPES, \
    CopyNumberType, RelativeCopyClass
from variation.schemas.token_response_schema import Nomenclature
from variation.schemas.validation_response_schema import ValidationSummary
from variation.classifiers import Classify
from variation.tokenizers import Tokenize
from variation.validators import Validate
from variation.translators import Translate
from variation.tokenizers import GeneSymbol
from variation.tokenizers.caches import AminoAcidCache
from variation.schemas.normalize_response_schema\
    import HGVSDupDelMode as HGVSDupDelModeEnum


class ToVRS:
    """The class for translating variation strings to VRS representations."""

    def __init__(self, tokenizer: Tokenize, classifier: Classify,
                 seqrepo_access: SeqRepoAccess,
                 transcript_mappings: TranscriptMappings,
                 gene_symbol: GeneSymbol, amino_acid_cache: AminoAcidCache,
                 uta: UTADatabase, mane_transcript_mappings: MANETranscriptMappings,
                 mane_transcript: MANETranscript, validator: Validate,
                 translator: Translate,
                 gene_normalizer: GeneQueryHandler,
                 hgvs_dup_del_mode: HGVSDupDelMode) -> None:
        """Initialize the ToVRS class.

        :param Tokenize tokenizer: Tokenizer class for tokenizing
        :param Classify classifier: Classifier class for classifying tokens
        :param SeqRepoAccess seqrepo_access: Access to SeqRepo
        :param TranscriptMappings transcript_mappings: Transcript mappings
            data class
        :param GeneSymbol gene_symbol: Class for identifying gene symbols
        :param AminoAcidCache amino_acid_cache: Amino Acid data class
        :param UTADatabase uta: UTA DB and queries
        :param MANETranscriptMappings mane_transcript_mappings: Class for
            getting mane transcript data from gene
        :param MANETranscript mane_transcript: Mane transcript data class
        :param Validate validator: Validator class for validating valid inputs
        :param Translate translator: Translating valid inputs
        :param GeneQueryHandler gene_normalizer: Gene normalizer access
        :param HGVSDupDelMode hgvs_dup_del_mode: Class for handling
            HGVS dup/del expressions
        """
        self.tokenizer = tokenizer
        self.classifier = classifier
        self.seq_repo_access = seqrepo_access
        self.transcript_mappings = transcript_mappings
        self.gene_symbol = gene_symbol
        self.amino_acid_cache = amino_acid_cache
        self.uta = uta
        self.mane_transcript_mappings = mane_transcript_mappings
        self.mane_transcript = mane_transcript
        self.validator = validator
        self.translator = translator
        self.gene_normalizer = gene_normalizer
        self.hgvs_dup_del_mode = hgvs_dup_del_mode

    async def get_validations(
            self, q: str, endpoint_name: Optional[Endpoint] = None,
            hgvs_dup_del_mode: Optional[Union[HGVSDupDelModeEnum, CopyNumberType]] = None,  # noqa: E501
            baseline_copies: Optional[int] = None,
            relative_copy_class: Optional[RelativeCopyClass] = None,
            do_liftover: bool = False
    ) -> Tuple[Optional[ValidationSummary], Optional[List[str]]]:
        """Return validation results for a given variation.

        :param str q: variation to get validation results for
        :param Optional[Endpoint] endpoint_name: Then name of the endpoint being used
        :param Optional[Union[HGVSDupDelModeEnum, CopyNumberType]] hgvs_dup_del_mode:
            This parameter determines how to interpret HGVS dup/del expressions
            in VRS.
        :param Optional[int] baseline_copies: Baseline copies number
        :param Optional[RelativeCopyClass] relative_copy_class: The relative copy class
        :param bool do_liftover: Whether or not to liftover to GRCh38 assembly
        :return: ValidationSummary for the variation and list of warnings
        """
        warnings = list()
        if q is None:
            return None, ["No variation entered"]
        tokens = self.tokenizer.perform(unquote(q.strip()), warnings)

        # gnomad vcf should always be a literal seq expression (allele)
        nomenclature = {t.nomenclature for t in tokens}
        if Nomenclature.GNOMAD_VCF in nomenclature:
            hgvs_dup_del_mode = HGVSDupDelModeEnum.LITERAL_SEQ_EXPR
        else:
            if endpoint_name == Endpoint.NORMALIZE:
                if hgvs_dup_del_mode:
                    hgvs_dup_del_mode = hgvs_dup_del_mode.strip().lower()
                    if not self.hgvs_dup_del_mode.is_valid_dup_del_mode(hgvs_dup_del_mode):  # noqa: E501
                        warnings.append(
                            f"hgvs_dup_del_mode must be one of: "
                            f"{self.hgvs_dup_del_mode.valid_dup_del_modes}")
                        return None, warnings
                else:
                    hgvs_dup_del_mode = HGVSDupDelModeEnum.DEFAULT
            elif endpoint_name in [Endpoint.HGVS_TO_ABSOLUTE_CN,
                                   Endpoint.HGVS_TO_RELATIVE_CN]:
                if not hgvs_dup_del_mode:
                    warnings.append(f"hgvs_dup_del_mode must be either "
                                    f"{CopyNumberType.ABSOLUTE.value} or "
                                    f"{CopyNumberType.RELATIVE.value}")
                    return None, warnings
                if not self.hgvs_dup_del_mode.is_valid_copy_number_mode(hgvs_dup_del_mode):  # noqa: E501
                    warnings.append(f"hgvs_dup_del_mode must be one of "
                                    f"{self.hgvs_dup_del_mode.valid_copy_number_modes}")
                    return None, warnings
                if endpoint_name == Endpoint.HGVS_TO_RELATIVE_CN:
                    if not relative_copy_class:
                        warnings.append(f"{Endpoint.HGVS_TO_RELATIVE_CN} endpoint "
                                        f"requires `relative_copy_class`")
                        return None, warnings
            else:
                hgvs_dup_del_mode = HGVSDupDelModeEnum.DEFAULT

        classifications = self.classifier.perform(tokens)

        if endpoint_name in [Endpoint.HGVS_TO_ABSOLUTE_CN,
                             Endpoint.HGVS_TO_RELATIVE_CN]:
            tmp_classifications = []
            for c in classifications:
                conditions = (
                    c.classification_type in VALID_CLASSIFICATION_TYPES,
                    ("HGVS" in c.matching_tokens or "ReferenceSequence" in c.matching_tokens)  # noqa: E501
                )
                if all(conditions):
                    tmp_classifications.append(c)
            classifications = tmp_classifications
            if not classifications:
                warnings = [f"{q} is not a supported HGVS genomic "
                            f"duplication or deletion"]
                return None, warnings

        validations = await self.validator.perform(
            classifications, endpoint_name=endpoint_name, warnings=warnings,
            hgvs_dup_del_mode=hgvs_dup_del_mode, baseline_copies=baseline_copies,
            relative_copy_class=relative_copy_class, do_liftover=do_liftover
        )
        if not warnings:
            warnings = validations.warnings
        return validations, warnings

    def get_translations(self, validations: ValidationSummary,
                         warnings: List)\
            -> Tuple[Optional[Union[List[Allele], List[AbsoluteCopyNumber],
                                    List[Text], List[Haplotype],
                                    List[VariationSet]]],
                     Optional[List[str]]]:
        """Return a list translations from a ValidationSummary.

        :param ValidationSummary validations: Valid and Invalid results
        :param List warnings: List of warnings
        :return: A list of unique translations from valid results
        """
        translations = []
        if validations is not None:
            for valid_variation in validations.valid_results:
                result = self.translator.perform(valid_variation)
                if result not in translations:
                    translations.append(result)
            if not translations and not warnings:
                warnings.append("Unable to validate variation")
        return translations, warnings



"""Module for testing Coding DNA Deletion Classifier."""
import unittest

from variation.classifiers import CodingDNADeletionClassifier
from .classifier_base import ClassifierBase


class TestCodingDNADeletionClassifier(ClassifierBase, unittest.TestCase):
    """A class to test the Coding DNA Deletion Classifier."""

    def classifier_instance(self):
        """Return CodingDNADeletionClassifier instance."""
        return CodingDNADeletionClassifier()

    def fixture_name(self):
        """Return CodingDNADeletionClassifier fixture name."""
        return "coding_dna_deletion"



"""Module for Protein Deletion Translation."""
from variation.translators.translator import Translator
from variation.schemas.classification_response_schema import ClassificationType
from variation.schemas.token_response_schema import ProteinDeletionToken, Token


class ProteinDeletion(Translator):
    """The Protein Deletion Translator class."""

    def can_translate(self, type: ClassificationType) -> bool:
        """Return if classification type is Protein Deletion."""
        return type == ClassificationType.PROTEIN_DELETION

    def is_token_instance(self, token: Token) -> bool:
        """Return if the token is an Protein Deletion token instance."""
        return isinstance(token, ProteinDeletionToken)



"""Module for Protein Termination classification."""
from typing import List

from variation.schemas.classification_response_schema import ClassificationType
from .set_based_classifier import SetBasedClassifier


class ProteinTerminationClassifier(SetBasedClassifier):
    """The protein termination classifier class."""

    def classification_type(self) -> ClassificationType:
        """Return Protein Termination classification type."""
        return ClassificationType.PROTEIN_TERMINATION

    def exact_match_candidates(self) -> List[List[str]]:
        """Return list of tokens for protein termination classification."""
        return [
            ["GeneSymbol", "ProteinTermination"],
            ["ProteinTermination"]
        ]



"""A module for Protein Substitution Tokenization."""
from typing import List, Optional

from variation.schemas.token_response_schema import SilentMutationToken,\
    TokenMatchType
from .polypeptide_sequence_variation_base import \
    PolypeptideSequenceVariationBase


class SilentMutation(PolypeptideSequenceVariationBase):
    """Class for tokenizing Protein Substitution."""

    def match(self, input_string: str) -> Optional[SilentMutationToken]:
        """Return a ProteinSubstitutionToken match if one exists.

        :param str input_string: The input string to match
        :return: A ProteinSubstitutionToken if a match exists. Otherwise, None.
        """
        if input_string is None:
            return None

        input_string = str(input_string).lower()
        psub_parts = None

        self.psub = {
            "amino_acid": None,
            "position": None,
            "new_amino_acid": None
        }

        if "." in input_string:
            if not input_string.startswith("p."):
                return None
            p_count = input_string.count("p.")
            if p_count == 1:
                psub_parts = self.splitter.split(input_string)
        else:
            psub_parts = self.splitter.split(input_string)

        self._get_psub(psub_parts)

        if None not in self.psub.values():
            if self.psub["new_amino_acid"] == "=":
                if not self._is_valid_amino_acid({self.psub["amino_acid"]}):
                    return None

                return SilentMutationToken(
                    token=input_string,
                    input_string=input_string,
                    match_type=TokenMatchType.UNSPECIFIED.value,
                    ref_protein=self.psub["amino_acid"],
                    position=self.psub["position"]
                )

        return None

    def _get_psub(self, psub_parts: List) -> None:
        """Get silent mutation tokens.

        :param List psub_parts: The split input string
        """
        psub_parts_len = len(psub_parts)
        if psub_parts_len == 3:
            if "p." in psub_parts[0]:
                psub_parts[0] = psub_parts[0].split("p.")[-1]
            else:
                if not psub_parts[0] and psub_parts[1] and not psub_parts[2]:
                    return

            if "(" in psub_parts[0] and ")" in psub_parts[2]:
                psub_parts[0] = psub_parts[0].split("(")[-1]
                psub_parts[2] = psub_parts[2].split(")")[0]

            self._set_psub(psub_parts[0], psub_parts[1], psub_parts[2])
        elif psub_parts_len == 2:
            psub_parts[0] = psub_parts[0].split(":")[-1]
        elif psub_parts_len == 1:
            if "p." in psub_parts[0]:
                psub_parts[0] = psub_parts[0].split("p.")[1]



"""A module for the Polypeptide Truncation Classifier."""
from typing import List

from variation.schemas.classification_response_schema import ClassificationType
from .set_based_classifier import SetBasedClassifier


class PolypeptideTruncationClassifier(SetBasedClassifier):
    """The Polypeptide Truncation Classifier class."""

    def classification_type(self) -> ClassificationType:
        """Return the Polypeptide Truncation classification type."""
        return ClassificationType.POLYPEPTIDE_TRUNCATION

    def exact_match_candidates(self) -> List[List[str]]:
        """Return the exact match token type candidates."""
        return [
            ["PolypeptideTruncation"],
            ["GeneSymbol", "PolypeptideTruncation"],
            ["HGVS", "PolypeptideTruncation"],
            ["ReferenceSequence", "PolypeptideTruncation"]
        ]



"""Module for toVRS endpoint response schema."""
from typing import List, Dict, Type, Any, Optional, Union

from pydantic import BaseModel
from pydantic.types import StrictStr
from ga4gh.vrsatile.pydantic.vrs_models import Allele, Text, Haplotype, \
    AbsoluteCopyNumber, VariationSet

from variation.schemas.normalize_response_schema import ServiceMeta


class ToVRSService(BaseModel):
    """Define model for translation response."""

    search_term: StrictStr
    warnings: Optional[List[StrictStr]]
    variations: Optional[Union[List[Allele], List[Text], List[Haplotype],
                               List[AbsoluteCopyNumber], List[VariationSet]]] = None
    service_meta_: ServiceMeta

    class Config:
        """Configure model."""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type["ToVRSService"]) -> None:
            """Configure OpenAPI schema."""
            if "title" in schema.keys():
                schema.pop("title", None)
            for prop in schema.get("properties", {}).values():
                prop.pop("title", None)
            schema["example"] = {
                "search_term": "BRAF V600E",
                "variations": [
                    {
                        "_id": "ga4gh:VA.7ys8TiDzrk04O3Upd63__rOBCEhv3P5d",
                        "type": "Allele",
                        "location": {
                            "_id": "ga4gh:VSL.Vxqx2bv42rWeu08Eg7JpkdQkMCNLskoz",  # noqa: E501
                            "type": "SequenceLocation",
                            "sequence_id": "ga4gh:SQ.ZJwurRo2HLY018wghYjDKSfIlEH0Y8At",  # noqa: E501
                            "interval": {
                                "type": "SequenceInterval",
                                "start": {
                                    "type": "Number",
                                    "value": 599
                                },
                                "end": {
                                    "type": "Number",
                                    "value": 600
                                }
                            }
                        },
                        "state": {
                            "type": "LiteralSequenceExpression",
                            "sequence": "E"
                        }
                    },
                    {
                        "_id": "ga4gh:VA.vimwyw0pFTwatfFhi3rhhb153ARWsPrW",
                        "type": "Allele",
                        "location": {
                            "_id": "ga4gh:VSL.FVmsWpfSOA3B2ryq0k995oHMuSGiFvMa",  # noqa: E501
                            "type": "SequenceLocation",
                            "sequence_id": "ga4gh:SQ.lKdPZpuT-VNvRuKDjsUItNgutfWYgWQd",  # noqa: E501
                            "interval": {
                                "type": "SequenceInterval",
                                "start": {
                                    "type": "Number",
                                    "value": 599
                                },
                                "end": {
                                    "type": "Number",
                                    "value": 600
                                }
                            }
                        },
                        "state": {
                            "type": "LiteralSequenceExpression",
                            "sequence": "E"
                        }
                    },
                    {
                        "_id": "ga4gh:VA.FzlrH5feNcQ3S9GayMU9EF008j-8Pbz5",
                        "type": "Allele",
                        "location": {
                            "_id": "ga4gh:VSL.QDLST2nKpPWwIArdO57L2VIWPNZ0DiN3",  # noqa: E501
                            "type": "SequenceLocation",
                            "sequence_id": "ga4gh:SQ.0Q-SgJX1V3seUUIu3qVUtEa55CQsGmEU",  # noqa: E501
                            "interval": {
                                "type": "SequenceInterval",
                                "start": {
                                    "type": "Number",
                                    "value": 599
                                },
                                "end": {
                                    "type": "Number",
                                    "value": 600
                                }
                            }
                        },
                        "state": {
                            "type": "LiteralSequenceExpression",
                            "sequence": "E"
                        }
                    },
                    {
                        "_id": "ga4gh:VA.ZDdoQdURgO2Daj2NxLj4pcDnjiiAsfbO",
                        "type": "Allele",
                        "location": {
                            "_id": "ga4gh:VSL.2cHIgn7iLKk4x9z3zLkSTTFMV0e48DR4",  # noqa: E501
                            "type": "SequenceLocation",
                            "sequence_id": "ga4gh:SQ.cQvw4UsHHRRlogxbWCB8W-mKD4AraM9y",  # noqa: E501
                            "interval": {
                                "type": "SequenceInterval",
                                "start": {
                                    "type": "Number",
                                    "value": 599
                                },
                                "end": {
                                    "type": "Number",
                                    "value": 600
                                }
                            }
                        },
                        "state": {
                            "type": "LiteralSequenceExpression",
                            "sequence": "E"
                        }
                    }
                ],
                "service_meta_": {
                    "name": "variation-normalizer",
                    "version": "0.2.17",
                    "response_datetime": "2022-01-26T22:23:41.821673",
                    "url": "https://github.com/cancervariants/variation-normalization"  # noqa: E501
                }

            }



"""Module for Validation."""
import copy
from typing import List, Optional, Dict, Tuple
from abc import ABC, abstractmethod
import logging

from ga4gh.vrsatile.pydantic.vrs_models import RelativeCopyClass
from gene.query import QueryHandler as GeneQueryHandler
from ga4gh.vrs.dataproxy import SeqRepoDataProxy
from ga4gh.vrs.extras.translator import Translator
from uta_tools.data_sources import SeqRepoAccess, TranscriptMappings, MANETranscript, \
    UTADatabase

from variation.schemas.classification_response_schema import Classification, \
    ClassificationType
from variation.schemas.app_schemas import Endpoint
from variation.schemas.token_response_schema import GeneMatchToken, Token, \
    GenomicSubstitutionToken
from variation.schemas.validation_response_schema import ValidationResult
from variation.tokenizers import GeneSymbol
from variation.validators.genomic_base import GenomicBase
from variation.schemas.normalize_response_schema\
    import HGVSDupDelMode as HGVSDupDelModeEnum
from variation.vrs import VRS

logger = logging.getLogger("variation")
logger.setLevel(logging.DEBUG)


class Validator(ABC):
    """The validator class."""

    def __init__(self, seqrepo_access: SeqRepoAccess,
                 transcript_mappings: TranscriptMappings,
                 gene_symbol: GeneSymbol,
                 mane_transcript: MANETranscript,
                 uta: UTADatabase, dp: SeqRepoDataProxy, tlr: Translator,
                 gene_normalizer: GeneQueryHandler,
                 vrs: VRS) -> None:
        """Initialize the DelIns validator.

        :param SeqRepoAccess seqrepo_access: Access to SeqRepo data
        :param TranscriptMappings transcript_mappings: Access to transcript
            mappings
        :param GeneSymbol gene_symbol: Gene symbol tokenizer
        :param MANETranscript mane_transcript: Access MANE Transcript
            information
        :param UTADatabase uta: Access to UTA queries
        :param Translator tlr: Translator class
        :param GeneQueryHandler gene_normalizer: Access to gene-normalizer
        :param VRS vrs: Class for creating VRS objects
        """
        self.transcript_mappings = transcript_mappings
        self.seqrepo_access = seqrepo_access
        self._gene_matcher = gene_symbol
        self.dp = dp
        self.tlr = tlr
        self.uta = uta
        self.genomic_base = GenomicBase(self.dp, self.uta)
        self.mane_transcript = mane_transcript
        self.gene_normalizer = gene_normalizer
        self.vrs = vrs

    @abstractmethod
    def is_token_instance(self, t: Token) -> bool:
        """Check to see if token is instance of a token type.

        :param Token t: Classification token to find type of
        :return: `True` if token is instance of class token. `False` otherwise.
        """
        raise NotImplementedError

    @abstractmethod
    def variation_name(self) -> str:
        """Return the variation name.

        :return: variation class name
        """
        raise NotImplementedError

    @abstractmethod
    def get_gene_tokens(
            self, classification: Classification) -> List[GeneMatchToken]:
        """Return a list of gene tokens for a classification.

        :param Classification classification: Classification for a list of
            tokens
        :return: A list of gene tokens for the classification
        """
        raise NotImplementedError

    @abstractmethod
    async def get_transcripts(self, gene_tokens: List, classification: Classification,
                              errors: List) -> Optional[List[str]]:
        """Get transcript accessions for a given classification.

        :param List gene_tokens: A list of gene tokens
        :param Classification classification: A classification for a list of
            tokens
        :param List errors: List of errors
        :return: List of transcript accessions
        """
        raise NotImplementedError

    @abstractmethod
    def validates_classification_type(
            self, classification_type: ClassificationType) -> bool:
        """Check that classification type can be validated by validator.

        :param ClassificationType classification_type: The type of variation
        :return: `True` if classification_type matches validator's
            classification type. `False` otherwise.
        """
        raise NotImplementedError

    @abstractmethod
    async def get_valid_invalid_results(
        self, classification_tokens: List, transcripts: List,
        classification: Classification, results: List, gene_tokens: List,
        mane_data_found: Dict, is_identifier: bool,
        hgvs_dup_del_mode: HGVSDupDelModeEnum,
        endpoint_name: Optional[Endpoint] = None,
        baseline_copies: Optional[int] = None,
        relative_copy_class: Optional[RelativeCopyClass] = None,
        do_liftover: bool = False
    ) -> None:
        """Add validation result objects to a list of results.

        :param List classification_tokens: A list of classification Tokens
        :param List transcripts: A list of transcript accessions
        :param Classification classification: A classification for a list of
            tokens
        :param List results: Stores validation result objects
        :param List gene_tokens: List of GeneMatchTokens for a classification
        :param Dict mane_data_found: MANE Transcript information found
        :param bool is_identifier: `True` if identifier is given for exact
            location. `False` otherwise.
        :param HGVSDupDelModeEnum hgvs_dup_del_mode: Must be: `default`, `cnv`,
            `repeated_seq_expr`, `literal_seq_expr`.
            This parameter determines how to represent HGVS dup/del expressions
            as VRS objects.
        :param Optional[Endpoint] endpoint_name: Then name of the endpoint being used
        :param Optional[int] baseline_copies: Baseline copies number
        :param Optional[RelativeCopyClass] relative_copy_class: The relative copy class
        :param bool do_liftover: Whether or not to liftover to GRCh38 assembly
        """
        raise NotImplementedError

    async def validate(
            self, classification: Classification,
            hgvs_dup_del_mode: HGVSDupDelModeEnum = HGVSDupDelModeEnum.DEFAULT,
            endpoint_name: Optional[Endpoint] = None,
            baseline_copies: Optional[int] = None,
            relative_copy_class: Optional[RelativeCopyClass] = None,
            do_liftover: bool = False
    ) -> List[ValidationResult]:
        """Return validation result for a given classification.

        :param Classification classification: A classification for a list of
            tokens
        :param HGVSDupDelModeEnum hgvs_dup_del_mode: Must be: `default`, `cnv`,
            `repeated_seq_expr`, `literal_seq_expr`.
            This parameter determines how to represent HGVS dup/del expressions
            as VRS objects.
        :param Optional[Endpoint] endpoint_name: Then name of the endpoint being used
        :param Optional[int] baseline_copies: Baseline copies number
        :param Optional[RelativeCopyClass] relative_copy_class: The relative copy class
        :param bool do_liftover: Whether or not to liftover to GRCh38 assembly
        :return: List of ValidationResult's containing valid and invalid
            results
        """
        results = list()
        errors = list()

        classification_tokens = self.get_classification_tokens(classification)
        if len(classification.non_matching_tokens) > 0:
            errors.append(f"Non matching tokens found for "
                          f"{self.variation_name()}.")

        gene_tokens = self.get_gene_tokens(classification)
        if len(gene_tokens) > 1:
            errors.append("More than one gene symbol found for a single"
                          f" {self.variation_name()}")

        try:
            # NC_ queries do not have gene tokens
            transcripts = await self.get_transcripts(
                gene_tokens, classification, errors)
        except IndexError:
            transcripts = list()

        if len(errors) > 0:
            return [
                self.get_validation_result(
                    classification, None, False, 0, {}, errors, gene_tokens
                )
            ]

        mane_data_found = {
            "mane_select": dict(),
            "mane_plus_clinical": dict(),
            "longest_compatible_remaining": dict(),
            "grch38": dict()
        }

        # If is_identifier, should only run once
        if "HGVS" in classification.matching_tokens:
            is_identifier = True
        else:
            is_identifier = False

        await self.get_valid_invalid_results(
            classification_tokens, transcripts, classification,
            results, gene_tokens, mane_data_found,
            is_identifier, hgvs_dup_del_mode, endpoint_name, baseline_copies,
            relative_copy_class, do_liftover
        )
        return results

    @staticmethod
    def get_validation_result(
            classification: Classification, classification_token: Token,
            is_valid: bool, confidence_score: int, variation: Dict,
            errors: List, gene_tokens: List, identifier: str = None,
            is_mane_transcript: bool = False) -> ValidationResult:
        """Return a validation result object.

        :param Classification classification: The classification for tokens
        :param Token classification_token: Classification token
        :param bool is_valid: Whether or not the classification is valid
        :param int confidence_score: The classification confidence score
        :param Dict variation: A VRS Variation object
        :param List errors: A list of errors for the classification
        :param List gene_tokens: List of GeneMatchTokens
        :param str identifier: Identifier for variation
        :param bool is_mane_transcript: `True` if result is MANE transcript.
            `False` otherwise.
        :return: A validation result
        """
        return ValidationResult(
            classification=classification,
            classification_token=classification_token,
            is_valid=is_valid,
            confidence_score=confidence_score,
            variation=variation,
            errors=errors,
            gene_tokens=gene_tokens,
            is_mane_transcript=is_mane_transcript,
            identifier=identifier
        )

    def get_protein_transcripts(self, gene_tokens: List,
                                errors: List) -> Optional[List[str]]:
        """Get transcripts for variations with protein reference sequence.

        :param List gene_tokens: List of gene tokens for a classification
        :param List errors: List of errors
        :return: List of possible transcript accessions for the variation
        """
        transcripts = self.transcript_mappings.protein_transcripts(gene_tokens[0].token)
        if not transcripts:
            errors.append(f"No transcripts found for gene symbol "
                          f"{gene_tokens[0].token}")
        return transcripts

    def get_coding_dna_transcripts(self, gene_tokens: List,
                                   errors: List) -> Optional[List[str]]:
        """Get transcripts for variations with coding DNA reference sequence.

        :param List gene_tokens: List of gene tokens for a classification
        :param List errors: List of errors
        :return: List of possible transcript accessions for the variation
        """
        transcripts = self.transcript_mappings.coding_dna_transcripts(
            gene_tokens[0].token)
        if not transcripts:
            errors.append(f"No transcripts found for gene symbol "
                          f"{gene_tokens[0].token}")
        return transcripts

    async def get_genomic_transcripts(self, classification: Classification,
                                      errors: List) -> Optional[List[str]]:
        """Get NC accessions for variations with genomic reference sequence.

        :param Classification classification: Classification for a list of
            tokens
        :param List errors: List of errors
        :return: List of possible NC accessions for the variation
        """
        nc_accessions = await self.genomic_base.get_nc_accessions(classification)
        if not nc_accessions:
            errors.append("Could not find NC_ accession for "
                          f"{self.variation_name()}")
        return nc_accessions

    def get_classification_tokens(
            self, classification: Classification
    ) -> List[Optional[Classification]]:
        """Get classification tokens for a given instance.

        :param Classification classification: A classification for a list of
            tokens
        :return: A list of classification tokens
        """
        return [t for t in classification.all_tokens
                if self.is_token_instance(t)]

    @staticmethod
    def get_gene_symbol_tokens(
            classification: Classification) -> List[Optional[GeneMatchToken]]:
        """Return tokens with GeneSymbol token type from a classification.

        :param Classification classification: Classification of input string
        :return: List of Gene Match Tokens
        """
        return [t for t in classification.all_tokens
                if t.token_type == "GeneSymbol"]

    def _add_gene_symbol_to_tokens(self, gene_symbol: str, gene_symbols: List,
                                   gene_tokens: List) -> None:
        """Add a gene symbol to list of gene match tokens.

        :param str gene_symbol: Gene symbol
        :param List gene_symbols: List of gene symbols matched
        :param List gene_tokens: List of GeneMatchTokens
        """
        if gene_symbol and gene_symbol not in gene_symbols:
            gene_symbols.append(gene_symbol)
            gene_tokens.append(self._gene_matcher.match(
                gene_symbol))

    def _get_gene_tokens(self, classification: Classification,
                         mappings: List) -> List[Optional[GeneMatchToken]]:
        """Get gene symbol tokens for protein or transcript reference
        sequences.

        :param Classification classification: Classification for a list of
            tokens
        :param List mappings: List of transcript mapping methods for
            corresponding reference sequence
        :return: A list of gene match tokens
        """
        gene_tokens = self.get_gene_symbol_tokens(classification)
        if not gene_tokens:
            refseq = \
                ([t.token for t in classification.all_tokens if
                  t.token_type in ["HGVS", "ReferenceSequence",
                                   "LocusReferenceGenomic"]] or [None])[0]

            if not refseq:
                return []

            if ":" in refseq:
                refseq = refseq.split(":")[0]

            gene_symbols = list()
            for mapping in mappings:
                gene_symbol = mapping(refseq)
                self._add_gene_symbol_to_tokens(
                    gene_symbol, gene_symbols, gene_tokens
                )
                if gene_tokens:
                    break
        return gene_tokens

    def get_protein_gene_symbol_tokens(
            self, classification: Classification
    ) -> List[Optional[GeneMatchToken]]:
        """Return gene tokens for a classification with protein reference
        sequence.

        :param Classification classification: The classification for a list of
            tokens
        :return: A list of Gene Match Tokens in the classification
        """
        mappings = [
            self.transcript_mappings.get_gene_symbol_from_ensembl_protein,
            self.transcript_mappings.get_gene_symbol_from_refeq_protein,
            self.transcript_mappings.get_gene_symbol_from_lrg
        ]
        return self._get_gene_tokens(classification, mappings)

    def get_coding_dna_gene_symbol_tokens(
            self, classification: Classification
    ) -> List[Optional[GeneMatchToken]]:
        """Return gene symbol tokens for classifications with coding dna
        reference sequence.

        :param Classification classification: Classification of input string
        :return: A list of gene match tokens
        """
        mappings = [
            self.transcript_mappings.get_gene_symbol_from_refseq_rna,
            self.transcript_mappings.get_gene_symbol_from_ensembl_transcript,  # noqa: E501
            self.transcript_mappings.get_gene_symbol_from_lrg
        ]
        return self._get_gene_tokens(classification, mappings)

    @staticmethod
    def get_accession(t: str, classification: Classification) -> str:
        """Return accession for a classification

        :param str t: Accession
        :param Classification classification: Classification for token
        :return: Accession
        """
        if "HGVS" in classification.matching_tokens or \
                "ReferenceSequence" in classification.matching_tokens:
            hgvs_token = [t for t in classification.all_tokens if
                          isinstance(t, Token) and t.token_type
                          in ["HGVS", "ReferenceSequence"]][0]
            hgvs_expr = hgvs_token.input_string
            t = hgvs_expr.split(":")[0]
        return t

    def add_validation_result(
            self, variation: Dict, valid_variations: List, results: List,
            classification: Classification, s: Token, t: str,
            gene_tokens: List, errors: List, identifier: str = None,
            is_mane_transcript: bool = False) -> bool:
        """Add validation result to list of results.

        :param Dict variation: A VRS Variation object
        :param List valid_variations: A list containing current valid
            variations
        :param List results: A list of validation results
        :param Classification classification: The classification for tokens
        :param Token s: The classification token
        :param string t: Transcript
        :param List gene_tokens: List of GeneMatchTokens
        :param List errors: A list of errors for the classification
        :param str identifier: Identifier for variation
        :param bool is_mane_transcript: `True` if result is MANE transcript.
            `False` otherwise.
        """
        if not errors:
            if is_mane_transcript or \
                    (variation and variation not in valid_variations):
                results.append(
                    self.get_validation_result(
                        classification, s, True, 1, variation, [], gene_tokens,
                        identifier=identifier if identifier else t,
                        is_mane_transcript=is_mane_transcript
                    )
                )
                valid_variations.append(variation)
                return True
        else:
            results.append(
                self.get_validation_result(
                    classification, s, False, 1, variation, errors, gene_tokens,
                    identifier=identifier if identifier else t,
                    is_mane_transcript=is_mane_transcript
                )
            )
            return False

    async def _validate_gene_pos(self, gene: str, alt_ac: str, pos1: int, pos2: int,
                                 errors: List, pos3: int = None, pos4: int = None,
                                 residue_mode: str = "residue") -> None:
        """Validate whether free text genomic query is valid input.
        If invalid input, add error to list of errors

        :param str gene: Queried gene
        :param str alt_ac: Genomic accession
        :param int pos1: Queried genomic position
        :param int pos2: Queried genomic position
        :param int pos3: Queried genomic position
        :param int pos4: Queried genomic position
        :param str residue_mode: Must be either `inter-residue` or `residue`
        :param List errors: List of errors
        """
        gene_start_end = {"start": None, "end": None}
        resp = self.gene_normalizer.search(gene, incl="Ensembl")
        if resp.source_matches:
            ensembl_resp = resp.source_matches[0]
            if ensembl_resp.records[0].locations:
                ensembl_loc = ensembl_resp.records[0].locations[0]
                gene_start_end["start"] = ensembl_loc.interval.start.value
                gene_start_end["end"] = ensembl_loc.interval.end.value - 1

        if gene_start_end["start"] is None and gene_start_end["end"] is None:
            errors.append(f"gene-normalizer unable to find Ensembl location"
                          f"for {gene}")
        else:
            assembly = await self.uta.get_chr_assembly(alt_ac)
            if assembly:
                # Not in GRCh38 assembly. Gene normalizer only uses 38, so we
                # need to liftover to GRCh37 coords
                chromosome, assembly = assembly
                for key in gene_start_end.keys():
                    gene_pos = gene_start_end[key]
                    gene_pos_liftover = self.uta.liftover_38_to_37.convert_coordinate(  # noqa: E501
                        chromosome, gene_pos)
                    if gene_pos_liftover is None or len(gene_pos_liftover) == 0:
                        errors.append(f"{gene_pos} does not"
                                      f" exist on {chromosome}")
                        return None
                    else:
                        gene_start_end[key] = gene_pos_liftover[0][1]

            gene_start = gene_start_end["start"]
            gene_end = gene_start_end["end"]

            for pos in [pos1, pos2, pos3, pos4]:
                if pos not in ["?", None]:
                    if residue_mode == "residue":
                        pos -= 1
                    if not (gene_start <= pos <= gene_end):
                        errors.append(f"Position {pos} out of index on "
                                      f"{alt_ac} on gene, {gene}")

    @staticmethod
    def _get_coord_alt(coordinate: str, mane: Dict,
                       s_copy: Token) -> Optional[Tuple[str, str]]:
        """Get coordinate and alteration

        :param str coordinate: Coordinate used. Must be either `p`, `c`, or `g`
        :param Dict mane: Mane data
        :param Token s_copy: classification token
        :return: Coordinate, alteration
        """
        if coordinate == "g" and mane["status"].lower() != "grch38":
            s_copy.molecule_context = "transcript"
            s_copy.coordinate_type = "c"
            coordinate = s_copy.coordinate_type

            if isinstance(s_copy, GenomicSubstitutionToken) and \
                    mane["strand"] == "-":
                ref_rev = s_copy.ref_nucleotide[::-1]
                alt_rev = s_copy.new_nucleotide[::-1]

                complements = {
                    "A": "T",
                    "T": "A",
                    "C": "G",
                    "G": "C"
                }

                s_copy.ref_nucleotide = ""
                s_copy.new_nucleotide = ""
                for nt in ref_rev:
                    s_copy.ref_nucleotide += complements[nt]
                for nt in alt_rev:
                    s_copy.new_nucleotide += complements[nt]
                alt = s_copy.new_nucleotide
            else:
                alt = None
            return coordinate, alt
        return None

    def add_mane_data(
            self, mane: Dict, mane_data: Dict, coordinate: str, alt_type: str,
            s: Token, alt: str = None, mane_variation: Dict = None) -> None:
        """Add mane transcript information to mane_data.

        :param Dict mane: MANE data
        :param Dict mane_data: All MANE data found for given query
        :param str coordinate: Coordinate used. Must be either `p`, `c`, or `g`
        :param str alt_type: Type of alteration
        :param Token s: Classification token
        :param str alt: Alteration
        :param Dict mane_variation: VRS Variation for mane data
        """
        if not mane:
            return None

        s_copy = copy.deepcopy(s)
        coord_alt = self._get_coord_alt(coordinate, mane, s_copy)
        if coord_alt:
            coordinate = coord_alt[0] if coord_alt[0] else coordinate
            alt = coord_alt[1] if coord_alt[1] else alt
        if mane_variation is None:
            new_allele = self.vrs.to_vrs_allele(
                mane["refseq"], mane["pos"][0] + 1, mane["pos"][1] + 1,
                coordinate, alt_type, [],
                cds_start=mane.get("coding_start_site", None), alt=alt
            )
            variation = new_allele
        else:
            variation = mane_variation

        if not variation:
            return None

        self._add_dict_to_mane_data(mane["refseq"], s_copy, variation,
                                    mane_data, mane["status"])

    @staticmethod
    def _add_dict_to_mane_data(ac: str, s: Token, variation: Dict,
                               mane_data: Dict, status: str) -> None:
        """Add variation data to mane data for normalize endpoint.

        :param str ac: Accession
        :param Token s: Classification token
        :param Dict variation: VRS Variation object
        :param Dict mane_data: MANE Transcript data found for given query
        :param str status: Status for variation (GRCh38, MANE Select,
            MANE Clinical Plus)
        """
        _id = variation["_id"]
        key = "_".join(status.lower().split())

        if _id in mane_data[key].keys():
            mane_data[key][_id]["count"] += 1
        else:
            mane_data[key][_id] = {
                "classification_token": s,
                "accession": ac,
                "count": 1,
                "variation": variation,
                "label": ac  # TODO: Use VRS to translate
            }

    def add_mane_to_validation_results(
            self, mane_data: Dict, valid_alleles: List, results: List,
            classification: Classification, gene_tokens: List) -> None:
        """Add MANE Transcript data to list of validation results.

        :param Dict mane_data: MANE Transcript data found for given query
        :param List valid_alleles: A list containing current valid alleles
        :param List results: A list of validation results
        :param Classification classification: The classification for tokens
        :param List gene_tokens: List of GeneMatchTokens
        """
        mane_data_keys = mane_data.keys()
        for key in ["mane_select", "mane_plus_clinical", "grch38",
                    "longest_compatible_remaining"]:
            highest_count = 0
            mane_result = None
            mane_allele = None
            identifier = None
            if key not in mane_data_keys:
                continue
            _mane_data_keys = mane_data[key].keys()
            for mane_allele_id in _mane_data_keys:
                data = mane_data[key][mane_allele_id]

                if data["count"] > highest_count:
                    highest_count = data["count"]
                    mane_result = data
                    mane_allele = data["variation"]
                    identifier = data["accession"]

            if mane_allele:
                self.add_validation_result(
                    mane_allele, valid_alleles, results, classification,
                    mane_result["classification_token"],
                    mane_result["accession"], gene_tokens, [],
                    identifier=identifier, is_mane_transcript=True
                )
                return

    def _check_index(self, ac: str, pos: int, errors: List) -> Optional[str]:
        """Check that index actually exists

        :param str ac: Accession
        :param int pos: Position changes
        :param List errors: List of errors
        :return: Reference sequence
        """
        seq, w = self.seqrepo_access.get_reference_sequence(ac, pos)
        if not seq:
            errors.append(w)
        return seq

    @staticmethod
    def _grch38_dict(ac: str, pos: Tuple[int, int]) -> Dict:
        """Create dict for normalized concepts

        :param str ac: Acession
        :param Tuple[int, int] pos: Position changes
        :return: GRCh38 data
        """
        return dict(
            gene=None,
            refseq=ac if ac.startswith("NC") else None,
            ensembl=ac if ac.startswith("ENSG") else None,
            pos=pos,
            strand=None,
            status="GRCh38"
        )

    def _is_grch38_assembly(self, t: str) -> bool:
        """Return whether or not accession is GRCh38 assembly.

        :param str t: Accession
        :return: `True` if accession is GRCh38 assembly. `False` otherwise
        """
        return "GRCh38" in [a for a in self.dp.get_metadata(t)["aliases"] if a.startswith("GRCh")][0]  # noqa: E501



"""Module for testing gnomad_vcf_to_protein works correctly"""
import pytest
from ga4gh.vrsatile.pydantic.vrsatile_models import VariationDescriptor

from tests.conftest import assertion_checks


@pytest.fixture(scope="module")
def mmel1_gene_context():
    """Create MMEL1 gene context test fixture"""
    return {
        "id": "normalize.gene:HGNC%3A14668",
        "type": "GeneDescriptor",
        "label": "MMEL1",
        "xrefs": [
            "ensembl:ENSG00000142606",
            "ncbigene:79258"
        ],
        "alternate_labels": [
            "SEP",
            "NL1",
            "MMEL2",
            "MMEL1",
            "MELL1",
            "NL2",
            "NEPII",
            "NEP2"
        ],
        "extensions": [
            {
                "type": "Extension",
                "name": "symbol_status",
                "value": "approved"
            },
            {
                "type": "Extension",
                "name": "approved_name",
                "value": "membrane metalloendopeptidase like 1"
            },
            {
                "type": "Extension",
                "name": "chromosome_location",
                "value": {
                    "species_id": "taxonomy:9606",
                    "interval": {
                        "type": "CytobandInterval",
                        "start": "p36.32",
                        "end": "p36.32"
                    },
                    "_id": "ga4gh:VCL.euTdJgW7alHXcorcGDFtKB80BLicpDhy",
                    "type": "ChromosomeLocation",
                    "chr": "1"
                }
            },
            {
                "type": "Extension",
                "name": "associated_with",
                "value": [
                    "uniprot:Q495T6",
                    "ccds:CCDS30569",
                    "ucsc:uc001ajy.3",
                    "merops:M13.008",
                    "pubmed:11560781",
                    "pubmed:18539150",
                    "vega:OTTHUMG00000000846",
                    "orphanet:239881",
                    "refseq:NM_033467",
                    "ena.embl:AF336981",
                    "omim:618104"
                ]
            },
            {
                "type": "Extension",
                "name": "previous_symbols",
                "value": [
                    "MMEL2",
                    "MELL1",
                    "MMEL1"
                ]
            }
        ],
        "gene_id": "hgnc:14668"
    }


@pytest.fixture(scope="module")
def cdk11a_gene_context():
    """Create test fixture for CDK11A"""
    return {
        "id": "normalize.gene:HGNC%3A1730",
        "type": "GeneDescriptor",
        "label": "CDK11A",
        "xrefs": [
            "ensembl:ENSG00000008128",
            "ncbigene:728642"
        ],
        "alternate_labels": [
            "LOC100294398",
            "LOC100134433",
            "CDC2L3",
            "CDK11-p46",
            "CDC2L2",
            "CDK11-p58",
            "CDK11-p110",
            "p58GTA",
            "PITSLRE"
        ],
        "extensions": [
            {
                "type": "Extension",
                "name": "symbol_status",
                "value": "approved"
            },
            {
                "type": "Extension",
                "name": "approved_name",
                "value": "cyclin dependent kinase 11A"
            },
            {
                "type": "Extension",
                "name": "chromosome_location",
                "value": {
                    "species_id": "taxonomy:9606",
                    "interval": {
                        "type": "CytobandInterval",
                        "start": "p36.33",
                        "end": "p36.33"
                    },
                    "_id": "ga4gh:VCL.DPABdBwKwyUotcGmM7aFYFGQqIK4NpYr",
                    "type": "ChromosomeLocation",
                    "chr": "1"
                }
            },
            {
                "type": "Extension",
                "name": "associated_with",
                "value": [
                    "pubmed:9750192",
                    "pubmed:19884882",
                    "vega:OTTHUMG00000000703",
                    "ccds:CCDS44043",
                    "ccds:CCDS81254",
                    "ccds:CCDS44042",
                    "iuphar:1963",
                    "refseq:NM_024011",
                    "ucsc:uc009vks.4",
                    "ena.embl:AF067522",
                    "omim:116951",
                    "pubmed:7920654",
                    "uniprot:Q9UQ88",
                    "ccds:CCDS81253"
                ]
            },
            {
                "type": "Extension",
                "name": "previous_symbols",
                "value": [
                    "LOC100294398",
                    "LOC100134433",
                    "CDC2L2",
                    "CDC2L3"
                ]
            }
        ],
        "gene_id": "hgnc:1730"
    }


@pytest.fixture(scope="module")
def lrp8_gene_context():
    """Create LRP8 gene context test fixture"""
    return {
        "id": "normalize.gene:HGNC%3A6700",
        "type": "GeneDescriptor",
        "label": "LRP8",
        "xrefs": [
            "ensembl:ENSG00000157193",
            "ncbigene:7804"
        ],
        "alternate_labels": [
            "APOER2",
            "LRP-8",
            "HSZ75190",
            "LRP8",
            "MIPS",
            "MCI1"
        ],
        "extensions": [
            {
                "type": "Extension",
                "name": "symbol_status",
                "value": "approved"
            },
            {
                "type": "Extension",
                "name": "approved_name",
                "value": "LDL receptor related protein 8"
            },
            {
                "type": "Extension",
                "name": "chromosome_location",
                "value": {
                    "species_id": "taxonomy:9606",
                    "interval": {
                        "type": "CytobandInterval",
                        "start": "p32.3",
                        "end": "p32.3"
                    },
                    "_id": "ga4gh:VCL.2z2vSUsZ2_Kh1RdVbIU5_tkwlZW-pBgN",
                    "type": "ChromosomeLocation",
                    "chr": "1"
                }
            },
            {
                "type": "Extension",
                "name": "associated_with",
                "value": [
                    "ucsc:uc001cvi.4",
                    "ccds:CCDS30720",
                    "ccds:CCDS579",
                    "uniprot:Q14114",
                    "ccds:CCDS580",
                    "vega:OTTHUMG00000008924",
                    "pubmed:9079678",
                    "pubmed:8626535",
                    "ccds:CCDS578",
                    "refseq:NM_004631",
                    "ena.embl:D50678",
                    "omim:602600"
                ]
            },
            {
                "type": "Extension",
                "name": "previous_symbols",
                "value": [
                    "HSZ75190",
                    "MIPS",
                    "LRP8"
                ]
            }
        ],
        "gene_id": "hgnc:6700"
    }


@pytest.fixture(scope="module")
def atad3a_gene_context():
    """Create ATAD3A gene context test fixture"""
    return {
        "id": "normalize.gene:HGNC%3A25567",
        "type": "GeneDescriptor",
        "label": "ATAD3A",
        "xrefs": [
            "ensembl:ENSG00000197785",
            "ncbigene:55210"
        ],
        "alternate_labels": [
            "HAYOS",
            "FLJ10709",
            "PHRINL"
        ],
        "extensions": [
            {
                "type": "Extension",
                "name": "symbol_status",
                "value": "approved"
            },
            {
                "type": "Extension",
                "name": "approved_name",
                "value": "ATPase family AAA domain containing 3A"
            },
            {
                "type": "Extension",
                "name": "chromosome_location",
                "value": {
                    "species_id": "taxonomy:9606",
                    "interval": {
                        "type": "CytobandInterval",
                        "start": "p36.33",
                        "end": "p36.33"
                    },
                    "_id": "ga4gh:VCL.DPABdBwKwyUotcGmM7aFYFGQqIK4NpYr",
                    "type": "ChromosomeLocation",
                    "chr": "1"
                }
            },
            {
                "type": "Extension",
                "name": "associated_with",
                "value": [
                    "ccds:CCDS53259",
                    "ucsc:uc001aga.3",
                    "ena.embl:AK025865",
                    "vega:OTTHUMG00000000575",
                    "ccds:CCDS31",
                    "refseq:NM_018188",
                    "omim:612316",
                    "ccds:CCDS53260",
                    "orphanet:469968",
                    "pubmed:28158749",
                    "uniprot:Q9NVI7"
                ]
            }
        ],
        "gene_id": "hgnc:25567"
    }


@pytest.fixture(scope="module")
def mmel1_l30m(mmel1_gene_context):
    """Create test fixture for MMEL1 L30M"""
    params = {
        "id": "normalize.variation:1-2629397-G-T",
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA.9TFIzZ8M_qfiAfZTO1QvZ6toS4I_j4Z9",
        "variation": {
            "_id": "ga4gh:VA.9TFIzZ8M_qfiAfZTO1QvZ6toS4I_j4Z9",
            "location": {
                "_id": "ga4gh:VSL.0h5OViQETz4k5QePhKFiQJE038yy4jpS",
                "interval": {
                    "end": {"value": 30, "type": "Number"},
                    "start": {"value": 29, "type": "Number"},
                    "type": "SequenceInterval"
                },
                "sequence_id": "ga4gh:SQ.iQ8F_pnsiQOLohiV2qh3OWRZiftUt8jZ",
                "type": "SequenceLocation"
            },
            "state": {
                "sequence": "M",
                "type": "LiteralSequenceExpression"
            },
            "type": "Allele"
        },
        "molecule_context": "protein",
        "structural_type": "SO:0001606",
        "vrs_ref_allele_seq": "L",
        "gene_context": mmel1_gene_context
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="module")
def cdk11a_e314del(cdk11a_gene_context):
    """Create test fixture for CDK11A Glu314del"""
    params = {
        "id": "normalize.variation:1-1708855-TTCC-T",
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA.6GUBUCcortzNbx7Q5DMPzxiFTnvPowhi",
        "variation": {
            "_id": "ga4gh:VA.6GUBUCcortzNbx7Q5DMPzxiFTnvPowhi",
            "location": {
                "_id": "ga4gh:VSL.tnCPFL4wpErBLA6rnIcFrRUGpaO1639_",
                "interval": {
                    "end": {"value": 321, "type": "Number"},
                    "start": {"value": 308, "type": "Number"},
                    "type": "SequenceInterval"
                },
                "sequence_id": "ga4gh:SQ.N728VSRRMHJ1SrhJgKqJOCaa3l5Z4sqm",
                "type": "SequenceLocation"
            },
            "state": {
                "sequence": "EEEEEEEEEEEE",
                "type": "LiteralSequenceExpression"
            },
            "type": "Allele"
        },
        "molecule_context": "protein",
        "structural_type": "SO:0001604",
        "vrs_ref_allele_seq": "EEEEEEEEEEEEE",
        "gene_context": cdk11a_gene_context
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="module")
def protein_insertion2(lrp8_gene_context):
    """Create test fixture for LRP8 p.Gln25_Leu26insArg"""
    params = {
        "id": "normalize.variation:1-53327836-A-ACGC",
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA.O1MlG0mAJcxpAfd-dRZ_ZXhMdbm3OHHx",
        "variation": {
            "_id": "ga4gh:VA.O1MlG0mAJcxpAfd-dRZ_ZXhMdbm3OHHx",
            "location": {
                "_id": "ga4gh:VSL.5jPV69Go1DH3XWb1bbRpoiz-GeEtIxYv",
                "interval": {
                    "end": {"value": 25, "type": "Number"},
                    "start": {"value": 25, "type": "Number"},
                    "type": "SequenceInterval"
                },
                "sequence_id": "ga4gh:SQ.qgIh8--4F6IpxRwX_lVtD2BhepH5B5Ef",
                "type": "SequenceLocation"
            },
            "state": {
                "sequence": "R",
                "type": "LiteralSequenceExpression"
            },
            "type": "Allele"
        },
        "molecule_context": "protein",
        "structural_type": "SO:0001605",
        "gene_context": lrp8_gene_context
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="module")
def atad3a_loc():
    """Create test fixture for ATAD3A location"""
    return {
        "_id": "ga4gh:VSL.h7pLvn-VyN4H7GT0vBj6XD5PENOocxOR",
        "interval": {
            "end": {"value": 7, "type": "Number"},
            "start": {"value": 6, "type": "Number"},
            "type": "SequenceInterval"
        },
        "sequence_id": "ga4gh:SQ.MHPOY_7fv8V9SktyvaTxulVFSK6XCxM8",
        "type": "SequenceLocation"
    }


@pytest.fixture(scope="module")
def atad3a_i7v(atad3a_gene_context, atad3a_loc):
    """Create test fixture for ATAD3A Ile7Val"""
    params = {
        "id": "normalize.variation:1-1512287-A-G",
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA.T4StYcC72ctAMe5FGzI9XkyYbLr6VxVk",
        "variation": {
            "_id": "ga4gh:VA.T4StYcC72ctAMe5FGzI9XkyYbLr6VxVk",
            "location": atad3a_loc,
            "state": {
                "sequence": "V",
                "type": "LiteralSequenceExpression"
            },
            "type": "Allele"
        },
        "molecule_context": "protein",
        "structural_type": "SO:0001606",
        "vrs_ref_allele_seq": "I",
        "gene_context": atad3a_gene_context
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="module")
def atad3a_i7t(atad3a_gene_context, atad3a_loc):
    """Create test fixture for ATAD3A Ile7Thr"""
    params = {
        "id": "normalize.variation:1-1512288-T-C",
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA.7ohTnL0ULPFaOESzuh_SIRYgp9zGch_Z",
        "variation": {
            "_id": "ga4gh:VA.7ohTnL0ULPFaOESzuh_SIRYgp9zGch_Z",
            "location": atad3a_loc,
            "state": {
                "sequence": "T",
                "type": "LiteralSequenceExpression"
            },
            "type": "Allele"
        },
        "molecule_context": "protein",
        "structural_type": "SO:0001606",
        "vrs_ref_allele_seq": "I",
        "gene_context": atad3a_gene_context
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="module")
def atad3a_i7m(atad3a_gene_context, atad3a_loc):
    """Create test fixture for ATAD3A Ile7Met"""
    params = {
        "id": "normalize.variation:1-1512289-T-G",
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA.ua1plOEkzNo23uVhnbSYBSM0jS7to6tN",
        "variation": {
            "_id": "ga4gh:VA.ua1plOEkzNo23uVhnbSYBSM0jS7to6tN",
            "location": atad3a_loc,
            "state": {
                "sequence": "M",
                "type": "LiteralSequenceExpression"
            },
            "type": "Allele"
        },
        "molecule_context": "protein",
        "structural_type": "SO:0001606",
        "vrs_ref_allele_seq": "I",
        "gene_context": atad3a_gene_context
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="session")
def braf_v600l(braf_gene_context, braf_600loc):
    """Create test fixture for BRAF Val600Leu."""
    params = {
        "id": "normalize.variation:7-140753337-C-A",
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA.Ktev5asCsmUbHaQG6N-CdSp_g5FyJxLN",
        "variation": {
            "_id": "ga4gh:VA.Ktev5asCsmUbHaQG6N-CdSp_g5FyJxLN",
            "location": braf_600loc,
            "state": {
                "sequence": "L",
                "type": "LiteralSequenceExpression"
            },
            "type": "Allele"
        },
        "molecule_context": "protein",
        "structural_type": "SO:0001606",
        "vrs_ref_allele_seq": "V",
        "gene_context": braf_gene_context
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="session")
def braf_600_silent_mutation(braf_gene_context, braf_600loc):
    """Create test fixture for BRAF Val600=."""
    params = {
        "id": "normalize.variation:7-140753335-C-A",
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA.YUPFxYUZlYnr7Q4nIPiLV5BJznzF3YKl",
        "variation": {
            "_id": "ga4gh:VA.YUPFxYUZlYnr7Q4nIPiLV5BJznzF3YKl",
            "location": braf_600loc,
            "state": {
                "sequence": "V",
                "type": "LiteralSequenceExpression"
            },
            "type": "Allele"
        },
        "molecule_context": "protein",
        "structural_type": "SO:0001606",
        "vrs_ref_allele_seq": "V",
        "gene_context": braf_gene_context
    }
    return VariationDescriptor(**params)


@pytest.mark.asyncio
async def test_substitution(test_query_handler, braf_v600e, braf_v600l,
                            braf_600_silent_mutation, mmel1_l30m, atad3a_i7v,
                            atad3a_i7t, atad3a_i7m):
    """Test that substitution queries return correct response"""
    # Reading Frame 1, Negative Strand
    resp, w = await test_query_handler.gnomad_vcf_to_protein("7-140753337-C-A")
    assertion_checks(resp, braf_v600l, "7-140753337-C-A", ignore_id=True)
    assert w == []

    # Reading Frame 2, Negative Strand
    resp, w = await test_query_handler.gnomad_vcf_to_protein("7-140753336-A-T")
    assertion_checks(resp, braf_v600e, "7-140753336-A-T", ignore_id=True)
    assert w == []

    # Reading Frame 3, Negative Strand
    resp, w = await test_query_handler.gnomad_vcf_to_protein("7-140753335-C-A")
    assertion_checks(resp, braf_600_silent_mutation, "7-140753335-C-A", ignore_id=True)
    assert w == []

    # Reading Frame 3, Negative Strand
    resp, w = await test_query_handler.gnomad_vcf_to_protein("1-2629397-G-T")
    assertion_checks(resp, mmel1_l30m, "1-2629397-G-T")
    assert w == []

    # Reading Frame 1, Positive Strand
    resp, w = await test_query_handler.gnomad_vcf_to_protein("1-1512287-A-G")
    assertion_checks(resp, atad3a_i7v, "1-1512287-A-G")
    assert w == []

    # Reading Frame 2, Positive Strand
    resp, w = await test_query_handler.gnomad_vcf_to_protein("1-1512288-T-C")
    assertion_checks(resp, atad3a_i7t, "1-1512288-T-C")
    assert w == []

    # Reading Frame 3, Positive Strand
    resp, w = await test_query_handler.gnomad_vcf_to_protein("1-1512289-T-G")
    assertion_checks(resp, atad3a_i7m, "1-1512289-T-G", ignore_id=True)
    assert w == []


@pytest.mark.asyncio
async def test_silent_mutation(test_query_handler, vhl_silent):
    """Test that silent queries return correct response"""
    resp, w = await test_query_handler.gnomad_vcf_to_protein("3-10183714-C-C")
    assertion_checks(resp, vhl_silent, "3-10183714-C-C", ignore_id=True)
    assert w == []


@pytest.mark.asyncio
async def test_insertion(test_query_handler, protein_insertion,
                         protein_insertion2):
    """Test that insertion queries return correct response"""
    resp, w = await test_query_handler.gnomad_vcf_to_protein("7-55181319-C-CGGGTTG")
    assertion_checks(resp, protein_insertion, "7-55181319-C-CGGGTTG", ignore_id=True)
    assert w == []

    resp, w = await test_query_handler.gnomad_vcf_to_protein("1-53327836-A-ACGC")
    assertion_checks(resp, protein_insertion2, "1-53327836-A-ACGC")
    assert w == []


@pytest.mark.asyncio
async def test_deletion(test_query_handler, protein_deletion_np_range,
                        cdk11a_e314del):
    """Test that deletion queries return correct response"""
    resp, w = await test_query_handler.gnomad_vcf_to_protein(
        "17-39723966-TTGAGGGAAAACACAT-T")
    assertion_checks(resp, protein_deletion_np_range,
                     "17-39723966-TTGAGGGAAAACACAT-T", ignore_id=True)
    assert w == []

    resp, w = await test_query_handler.gnomad_vcf_to_protein("1-1708855-TTCC-T")
    assertion_checks(resp, cdk11a_e314del, "1-1708855-TTCC-T")
    assert w == []


@pytest.mark.asyncio
async def test_invalid(test_query_handler):
    """Test that invalid queries return correct response"""
    resp, w = await test_query_handler.gnomad_vcf_to_protein("BRAF V600E")
    assert resp.variation.type == "Text"
    assert resp.label == "BRAF V600E"
    assert w == ["BRAF V600E is not a supported gnomad vcf query"]

    resp, w = await test_query_handler.gnomad_vcf_to_protein("7-140753336-T-G")
    assert resp.variation.type == "Text"
    assert resp.label == "7-140753336-T-G"
    assert set(w) == {
        "Unable to get transcripts given NC_000007.13 and 140753336",
        "Expected T but found A on NC_000007.14 at position 140753336",
    }

    resp, w = await test_query_handler.gnomad_vcf_to_protein("20-2-TC-TG")
    assert resp.variation.type == "Text"
    assert resp.label == "20-2-TC-TG"
    assert w == ["Unable to get protein variation for 20-2-TC-TG"]



"""A module for the Silent Mutation Classifier."""
from typing import List

from variation.schemas.classification_response_schema import ClassificationType
from .set_based_classifier import SetBasedClassifier


class SilentMutationClassifier(SetBasedClassifier):
    """The Silent Mutation Classifier class."""

    def classification_type(self) -> ClassificationType:
        """Return the Silent Mutation classification type."""
        return ClassificationType.SILENT_MUTATION

    def exact_match_candidates(self) -> List[List[str]]:
        """Return the exact match token type candidates."""
        return [
            ["SilentMutation"],
            ["GeneSymbol", "SilentMutation"],
            ["HGVS", "SilentMutation"],
            ["ReferenceSequence", "SilentMutation"]
        ]



"""Main application for FastAPI."""
from typing import Dict, Optional
from datetime import datetime
from urllib.parse import unquote

import pkg_resources
from fastapi import FastAPI, Query
from fastapi.openapi.utils import get_openapi
import python_jsonschema_objects
from ga4gh.vrsatile.pydantic.vrs_models import RelativeCopyClass
from hgvs.exceptions import HGVSError
from bioutils.exceptions import BioutilsError

from variation.schemas import ToVRSService, NormalizeService, ServiceMeta
from variation.schemas.hgvs_to_copy_number_schema import \
    HgvsToAbsoluteCopyNumberService, HgvsToRelativeCopyNumberService
from variation.query import QueryHandler
from variation.schemas.normalize_response_schema \
    import HGVSDupDelMode as HGVSDupDelModeEnum, TranslateIdentifierService, \
    CanonicalSPDIToCategoricalVariationService
from .version import __version__
from .schemas.vrs_python_translator_schema import TranslateFromFormat, \
    TranslateFromService, TranslateFromQuery, VrsPythonMeta


app = FastAPI(
    docs_url="/variation",
    openapi_url="/variation/openapi.json",
    swagger_ui_parameters={"tryItOutEnabled": True}
)
query_handler = QueryHandler()


def custom_openapi() -> Dict:
    """Generate custom fields for OpenAPI response."""
    if app.openapi_schema:
        return app.openapi_schema
    openapi_schema = get_openapi(
        title="The VICC Variation Normalizer",
        version=__version__,
        description="Services and guidelines for normalizing variations.",
        routes=app.routes
    )

    openapi_schema["info"]["contact"] = {
        "name": "Alex H. Wagner",
        "email": "Alex.Wagner@nationwidechildrens.org",
        "url": "https://www.nationwidechildrens.org/specialties/institute-for-genomic-medicine/research-labs/wagner-lab"  # noqa: E501
    }
    app.openapi_schema = openapi_schema
    return app.openapi_schema


app.openapi = custom_openapi

translate_summary = "Translate a variation to a VRS compatible object."
translate_description = ("Translate a variation into Variation Representation "
                         "Specification (VRS) compatible representations.")
translate_response_description = "A  response to a validly-formed query."
q_description = "Variation to translate."


@app.get("/variation/toVRS",
         summary=translate_summary,
         response_description=translate_response_description,
         response_model=ToVRSService,
         description=translate_description,
         response_model_exclude_none=True
         )
async def to_vrs(q: str = Query(..., description=q_description)) -> ToVRSService:
    """Return a VRS-like representation of all validated variations for the search term.  # noqa: E501, D400

    :param str q: The variation to translate
    :return: ToVRSService model for variation
    """
    translations, warnings = await query_handler.to_vrs(unquote(q))

    return ToVRSService(
        search_term=q,
        variations=translations,
        service_meta_=ServiceMeta(
            version=__version__,
            response_datetime=datetime.now()
        ),
        warnings=warnings
    )


normalize_summary = "Given variation, return VRSATILE compatible object."
normalize_response_description = "A response to a validly-formed query."
normalize_description = \
    "Return VRSATILE compatible object for variation provided."
q_description = "Variation to normalize."
hgvs_dup_del_mode_decsr = ("Must be one of: `default`, `cnv`, "
                           "`repeated_seq_expr`, `literal_seq_expr`. This"
                           " parameter determines how to interpret HGVS "
                           "dup/del expressions in VRS.")


@app.get("/variation/normalize",
         summary=normalize_summary,
         response_description=normalize_response_description,
         response_model=NormalizeService,
         description=normalize_description,
         response_model_exclude_none=True
         )
async def normalize(
    q: str = Query(..., description=q_description),
    hgvs_dup_del_mode: Optional[HGVSDupDelModeEnum] = Query(
        None, description=hgvs_dup_del_mode_decsr)
) -> NormalizeService:
    """Return Value Object Descriptor for variation.

    :param str q: Variation to normalize
    :param Optional[HGVSDupDelModeEnum] hgvs_dup_del_mode:
        Must be: `default`, `cnv`, `repeated_seq_expr`, `literal_seq_expr`.
        This parameter determines how to interpret HGVS dup/del expressions
        in VRS.
    :return: NormalizeService for variation
    """
    normalize_resp = await query_handler.normalize(
        unquote(q), hgvs_dup_del_mode=hgvs_dup_del_mode)
    warnings = query_handler.normalize_handler.warnings if \
        query_handler.normalize_handler.warnings else None

    return NormalizeService(
        variation_query=q,
        variation_descriptor=normalize_resp,
        warnings=warnings,
        service_meta_=ServiceMeta(
            version=__version__,
            response_datetime=datetime.now()
        )
    )


@app.get("/variation/translate_identifier",
         summary="Given an identifier, use SeqRepo to return a list of aliases.",  # noqa: E501
         response_description="A response to a validly-formed query.",
         response_model=TranslateIdentifierService,
         description="Return list of aliases for an identifier"
         )
def translate_identifier(
        identifier: str = Query(..., description="The identifier to find aliases for"),  # noqa: E501
        target_namespaces: Optional[str] = Query(None, description="The namespaces of the aliases, separated by commas")  # noqa: E501
) -> TranslateIdentifierService:
    """Return data containing identifier aliases.

    :param str identifier: The identifier to find aliases for
    :param Optional[str] target_namespaces: The namespaces of the aliases,
        separated by commas
    :return: TranslateIdentifierService data
    """
    aliases = []
    warnings = None
    identifier = identifier.strip()
    try:
        aliases = query_handler.seqrepo_access.seqrepo_client.translate_identifier(  # noqa: E501
            identifier, target_namespaces=target_namespaces)
    except KeyError:
        warnings = [f"Identifier, {identifier}, does not exist in SeqRepo"]
    except Exception as e:
        warnings = [f"SeqRepo could not translate identifier, {identifier}:"
                    f" {e}"]

    return TranslateIdentifierService(
        identifier_query=identifier,
        warnings=warnings,
        aliases=aliases,
        service_meta_=ServiceMeta(
            version=__version__,
            response_datetime=datetime.now()
        ))


g_to_p_summary = "Given gnomad VCF, return VRSATILE compatible object on " \
                 "protein coordinate."
g_to_p_response_description = "A response to a validly-formed query."
g_to_p_description = \
    "Return VRSATILE compatible object on protein coordinate for " \
    "variation provided."
q_description = "gnomad VCF to normalize to protein variation."


@app.get("/variation/gnomad_vcf_to_protein",
         summary=g_to_p_summary,
         response_description=g_to_p_response_description,
         description=g_to_p_description,
         response_model=NormalizeService,
         response_model_exclude_none=True
         )
async def gnomad_vcf_to_protein(
    q: str = Query(..., description=q_description)
) -> NormalizeService:
    """Return Value Object Descriptor for variation on protein coordinate.

    :param str q: gnomad VCF to normalize to protein variation.
    :return: NormalizeService for variation
    """
    q = unquote(q.strip())
    resp, warnings = await query_handler.gnomad_vcf_to_protein(q)

    return NormalizeService(
        variation_query=q,
        variation_descriptor=resp,
        warnings=warnings,
        service_meta_=ServiceMeta(
            version=__version__,
            response_datetime=datetime.now()
        )
    )


complement_descr = "This field indicates that a categorical variation is defined to " \
                   "include (false) or exclude (true) variation concepts matching " \
                   "the categorical variation."


@app.get("/variation/canonical_spdi_to_categorical_variation",
         summary="Given canonical SPDI, return VRSATILE categorical variation object",
         response_description="A response to a validly-formed query.",
         description="Return VRSATILE categorical variation object",
         response_model=CanonicalSPDIToCategoricalVariationService,
         response_model_exclude_none=True)
def canonical_spdi_to_categorical_variation(
        q: str = Query(..., description="Canonical SPDI"),
        complement: bool = Query(False, description=complement_descr)
) -> CanonicalSPDIToCategoricalVariationService:
    """Return categorical variation for canonical SPDI

    :param str q: Canonical SPDI
    :param bool complement: This field indicates that a categorical variation
        is defined to include (false) or exclude (true) variation concepts matching the
        categorical variation. This is equivalent to a logical NOT operation on the
        categorical variation properties.
    :return: CanonicalSPDIToCategoricalVariationService for variation query
    """
    q = unquote(q.strip())
    resp, warnings = query_handler.canonical_spdi_to_categorical_variation(
        q, complement=complement)
    return CanonicalSPDIToCategoricalVariationService(
        canonical_spdi_query=q,
        categorical_variation=resp,
        warnings=warnings,
        service_meta_=ServiceMeta(
            version=__version__,
            response_datetime=datetime.now()
        )
    )


from_fmt_descr = "Format of input variation to translate. Must be one of `beacon`, " \
                 "`gnomad`, `hgvs`, or `spdi`"


@app.get("/variation/translate_from",
         summary="Given variation as beacon, gnomad, hgvs or spdi representation, "
                 "return VRS Allele object using vrs-python's translator class",
         response_description="A response to a validly-formed query.",
         description="Return VRS Allele object",
         response_model=TranslateFromService,
         response_model_exclude_none=True)
def vrs_python_translate_from(
    variation: str = Query(..., description="Variation to translate to VRS object."
                                            " Must be represented as either beacon, "
                                            "gnomad, hgvs, or spdi."),
    fmt: Optional[TranslateFromFormat] = Query(None, description=from_fmt_descr)
) -> TranslateFromService:
    """Given variation query, return VRS Allele object using vrs-python"s translator
        class

    :param str variation: Variation to translate to VRS object. Must be represented
        as either beacon, gnomad, hgvs, or spdi
    :param Optional[TranslateFromFormat] fmt: Format of variation. If not supplied,
        vrs-python will infer its format.
    :return: TranslateFromService containing VRS Allele object
    """
    variation_query = unquote(variation.strip())
    warnings = list()
    vrs_variation = None
    try:
        resp = query_handler.tlr.translate_from(variation_query, fmt)
    except (KeyError, ValueError, python_jsonschema_objects.validators.ValidationError) as e:  # noqa: E501
        warnings.append(f"vrs-python translator raised {type(e).__name__}: {e}")
    except HGVSError as e:
        warnings.append(f"hgvs raised {type(e).__name__}: {e}")
    except BioutilsError as e:
        warnings.append(f"bioutils raised {type(e).__name__}: {e}")
    else:
        vrs_variation = resp.as_dict()

    return TranslateFromService(
        query=TranslateFromQuery(variation=variation_query, fmt=fmt),
        warnings=warnings,
        variation=vrs_variation,
        service_meta_=ServiceMeta(
            version=__version__,
            response_datetime=datetime.now()
        ),
        vrs_python_meta_=VrsPythonMeta(
            version=pkg_resources.get_distribution("ga4gh.vrs").version
        )
    )


# @app.post("/variation/translate_to",
#           summary="Given VRS Allele object as a dict, return variation expressed as "
#                   "queried format using vrs-python"s translator class",
#           response_description="A response to a validly-formed query.",
#           description="Return variation in queried format representation. "
#                       "Request body must contain `allele` and `fmt`. `allele` is a "
#                       "VRS Allele object represented as a dict. `fmt` must be either "
#                       "`spdi` or `hgvs`",
#           response_model=TranslateToService,
#           response_model_exclude_none=True)
# async def vrs_python_translate_to(
#         request: Request) -> Union[ErrorResponse, TranslateToService]:
#     """Given VRS Allele object as a dict, return variation expressed as queried
#     format using vrs-python's translator class
#
#     :param Request request: Request body. Request body must contain `allele` and `fmt`.  # noqa
#         `allele` is a VRS Allele object represented as a dict. `fmt` must be either
#         `spdi` or `hgvs`
#     :return: ErrorResponse if invalid request body. Else, TranslateToService containing  # noqa
#         variation represented as fmt representation if valid VRS Allele
#     """
#     r = await request.json()
#     warnings = list()
#
#     allele_query = r.get("allele")
#     if not allele_query:
#         warnings.append("Missing `allele`. Must be VRS Allele represented as a dict")
#     else:
#         if not isinstance(allele_query, dict):
#             warnings.append("`allele` must be a dict")
#
#     if warnings:
#         return ErrorResponse(errors=warnings)
#
#     fmt_query = r.get("fmt")
#     if not fmt_query:
#         warnings.append("Missing `fmt`. Must be either `hgvs` or `spdi`")
#     else:
#         if not isinstance(fmt_query, str):
#             warnings.append("`fmt` must be a str")
#         else:
#             fmt_query = fmt_query.strip()
#             valid_fmts = [v.value for k, v in TranslateToFormat.__members__.items()]
#             if fmt_query not in valid_fmts:
#                 warnings.append(f"{fmt_query} not a valid fmt. "
#                                 f"Must be one of {valid_fmts}")
#
#     if warnings:
#         return ErrorResponse(errors=warnings)
#
#     allele = None
#     try:
#         allele = models.Allele(**r["allele"])
#     except ValidationError as e:
#         warnings.append(f"`allele` is not a valid VRS Allele: {e}")
#
#     variation = []
#     if allele:
#         try:
#             variation = query_handler.tlr.translate_to(allele, r["fmt"])
#         except ValueError as e:
#             warnings.append(f"vrs-python translator raised {type(e).__name__}: {e}")
#
#     return TranslateToService(
#         query=TranslateToQuery(variation=allele_query, fmt=fmt_query),
#         warnings=warnings,
#         variations=variation,
#         service_meta_=ServiceMeta(
#             version=__version__,
#             response_datetime=datetime.now()
#         ),
#         vrs_python_meta_=VrsPythonMeta(
#             version=pkg_resources.get_distribution("ga4gh.vrs").version
#         )
#     )

@app.get("/variation/hgvs_to_absolute_copy_number",
         summary="Given HGVS expression, return absolute copy number variation",
         response_description="A response to a validly-formed query.",
         description="Return VRS object",
         response_model=HgvsToAbsoluteCopyNumberService,
         response_model_exclude_none=True)
async def hgvs_to_absolute_copy_number(
    hgvs_expr: str = Query(..., description="Variation query"),
    baseline_copies: Optional[int] = Query(
        None, description="Baseline copies for duplication"),
    do_liftover: bool = Query(False, description="Whether or not to liftover "
                              "to GRCh38 assembly.")
) -> HgvsToAbsoluteCopyNumberService:
    """Given hgvs expression, return absolute copy number variation

    :param str hgvs_expr: HGVS expression
    :param Optional[int] baseline_copies: Baseline copies number
    :param bool do_liftover: Whether or not to liftover to GRCh38 assembly
    :return: HgvsToAbsoluteCopyNumberService
    """
    variation, warnings = await query_handler.hgvs_to_absolute_copy_number(
        unquote(hgvs_expr.strip()), baseline_copies, do_liftover)

    return HgvsToAbsoluteCopyNumberService(
        hgvs_expr=hgvs_expr,
        warnings=warnings,
        service_meta_=ServiceMeta(
            version=__version__,
            response_datetime=datetime.now()
        ),
        absolute_copy_number=variation
    )


@app.get("/variation/hgvs_to_relative_copy_number",
         summary="Given HGVS expression, return relative copy number variation",
         response_description="A response to a validly-formed query.",
         description="Return VRS object",
         response_model=HgvsToRelativeCopyNumberService,
         response_model_exclude_none=True)
async def hgvs_to_relative_copy_number(
    hgvs_expr: str = Query(..., description="Variation query"),
    relative_copy_class: RelativeCopyClass = Query(
        ..., description="The relative copy class"),
    do_liftover: bool = Query(False, description="Whether or not to liftover "
                              "to GRCh38 assembly.")
) -> HgvsToRelativeCopyNumberService:
    """Given hgvs expression, return relative copy number variation

    :param str hgvs_expr: HGVS expression
    :param RelativeCopyClass relative_copy_class: Relative copy class
    :param bool do_liftover: Whether or not to liftover to GRCh38 assembly
    :return: HgvsToRelativeCopyNumberService
    """
    variation, warnings = await query_handler.hgvs_to_relative_copy_number(
        unquote(hgvs_expr.strip()), relative_copy_class, do_liftover)

    return HgvsToRelativeCopyNumberService(
        hgvs_expr=hgvs_expr,
        warnings=warnings,
        service_meta_=ServiceMeta(
            version=__version__,
            response_datetime=datetime.now()
        ),
        relative_copy_number=variation
    )



"""Module for testing Protein Substitution Classifier."""
import unittest

from variation.classifiers import ProteinSubstitutionClassifier
from .classifier_base import ClassifierBase


class TestProteinSubstitutionClassifier(ClassifierBase, unittest.TestCase):
    """A class to test the Protein Substitution Classifier."""

    def classifier_instance(self):
        """Return ProteinSubstitutionClassifier instance."""
        return ProteinSubstitutionClassifier()

    def fixture_name(self):
        """Return ProteinSubstitutionClassifier fixture name."""
        return "protein_substitution"



"""Module for testing Coding DNA Deletion Translator."""
import unittest

from variation.classifiers import CodingDNADeletionClassifier
from variation.translators import CodingDNADeletion
from variation.validators import CodingDNADeletion as CDNAD_V
from .translator_base import TranslatorBase


class TestCodingDNADeletionTranslator(TranslatorBase, unittest.TestCase):
    """A class to test the Coding DNA Deletion Translator."""

    def classifier_instance(self):
        """Return coding DNA deletion instance."""
        return CodingDNADeletionClassifier()

    def validator_instance(self):
        """Return coding DNA delins instance."""
        return CDNAD_V(*self.params)

    def translator_instance(self):
        """Return coding DNA deletion instance."""
        return CodingDNADeletion()

    def fixture_name(self):
        """Return the fixture name for coding DNA deletion."""
        return "coding_dna_deletion"



"""A module for the Protein Deletion Classifier."""
from typing import List

from variation.schemas.classification_response_schema import ClassificationType
from .set_based_classifier import SetBasedClassifier


class ProteinDeletionClassifier(SetBasedClassifier):
    """The Protein Deletion Classifier class."""

    def classification_type(self) -> ClassificationType:
        """Return the Protein Deletion classification type."""
        return ClassificationType.PROTEIN_DELETION

    def exact_match_candidates(self) -> List[List[str]]:
        """Return the exact match token type candidates."""
        return [
            ["ProteinDeletion"],
            ["GeneSymbol", "ProteinSubstitution", "ProteinDeletion"],
            ["ProteinDeletion", "GeneSymbol"],
            ["GeneSymbol", "ProteinDeletion"],
            ["HGVS", "ProteinDeletion"],
            ["ReferenceSequence", "ProteinDeletion"],
            ["LocusReferenceGenomic", "ProteinDeletion"]
        ]



"""The Variation Normalization package."""
from pathlib import Path
import logging
from ftplib import FTP
from os import environ, remove
import gzip
import shutil

from .version import __version__  # noqa: F401


APP_ROOT = Path(__file__).resolve().parents[0]

if "VARIATION_NORM_EB_PROD" in environ:
    environ["GENE_NORM_EB_PROD"] = "true"
    LOG_FN = "/tmp/variation.log"
else:
    LOG_FN = "variation.log"

logging.basicConfig(
    filename=LOG_FN,
    format="[%(asctime)s] - %(name)s - %(levelname)s : %(message)s")
logger = logging.getLogger("variation")
logger.setLevel(logging.DEBUG)
logger.handlers = []

logging.getLogger("uta_tools").setLevel(logging.INFO)
logging.getLogger("boto3").setLevel(logging.INFO)
logging.getLogger("botocore").setLevel(logging.INFO)
logging.getLogger("urllib3").setLevel(logging.INFO)
logging.getLogger("python_jsonschema_objects").setLevel(logging.INFO)
logging.getLogger("hgvs.parser").setLevel(logging.INFO)
logging.getLogger("biocommons.seqrepo.seqaliasdb.seqaliasdb").setLevel(logging.INFO)
logging.getLogger("biocommons.seqrepo.fastadir.fastadir").setLevel(logging.INFO)

if "VARIATION_NORM_EB_PROD" in environ:
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    logger.addHandler(ch)


def data_download(path: Path, domain: str, dir: str, fn: str) -> None:
    """Download files using FTP.

    :param Path path: The path to the file
    :param str domain: The domain of the FTP site
    :param str dir: The directory that the file is located in
    :param str fn: The file name to download
    """
    if "VARIATION_NORM_EB_PROD" not in environ and not Path(path).exists():
        with FTP(domain) as ftp:
            ftp.login()
            ftp.cwd(dir)
            if fn.endswith(".gz") and not path.endswith(".gz"):
                path += ".gz"
            with open(f"{path}", "wb") as fp:
                ftp.retrbinary(f"RETR {fn}", fp.write)
            if fn.endswith(".gz"):
                with gzip.open(f"{path}", "rb") as f_in:
                    dest = path[:-3]
                    with open(dest, "wb") as f_out:
                        shutil.copyfileobj(f_in, f_out)
                remove(f"{path}")


AMINO_ACID_PATH = environ.get("AMINO_ACID_PATH", f"{APP_ROOT}/data/amino_acids.csv")
UTA_DB_URL = environ.get("UTA_DB_URL",
                         "postgresql://uta_admin@localhost:5433/uta/uta_20210129")



"""Module for Coding DNA Substitution Translation."""
from variation.translators.translator import Translator
from variation.schemas.classification_response_schema import ClassificationType
from variation.schemas.token_response_schema import CodingDNASubstitutionToken, Token


class CodingDNASubstitution(Translator):
    """The Coding DNA Substitution Translator class."""

    def can_translate(self, type: ClassificationType) -> bool:
        """Return if classification type is Coding DNA Substitution."""
        return type == ClassificationType.CODING_DNA_SUBSTITUTION

    def is_token_instance(self, token: Token) -> bool:
        """Return if the token is an Coding DNA Substitution token instance."""
        return isinstance(token, CodingDNASubstitutionToken)



"""Module for testing Genomic Duplication Translator."""
import unittest

from variation.classifiers import GenomicDuplicationClassifier
from variation.translators import GenomicDuplication
from variation.validators import GenomicDuplication as GD_V
from .translator_base import TranslatorBase


class TestGenomicDuplicationTranslator(TranslatorBase, unittest.TestCase):
    """A class to test the Genomic Duplication Translator."""

    def classifier_instance(self):
        """Return genomic duplication instance."""
        return GenomicDuplicationClassifier()

    def validator_instance(self):
        """Return genomic duplication instance."""
        return GD_V(*self.params)

    def translator_instance(self):
        """Return genomic duplication instance."""
        return GenomicDuplication()

    def fixture_name(self):
        """Return the fixture name for genomic duplication."""
        return "genomic_duplication"



"""Module for testing Coding DNA Silent Mutation Validator."""
import unittest

from variation.validators import CodingDNASilentMutation
from variation.classifiers import CodingDNASilentMutationClassifier
from .validator_base import ValidatorBase


class TestCodingDNASilentMutationValidator(ValidatorBase, unittest.TestCase):
    """A class to test the Coding DNA Silent Mutation Validator."""

    def validator_instance(self):
        """Return coding DNA silent mutation instance."""
        return CodingDNASilentMutation(*self.params)

    def classifier_instance(self):
        """Return the coding DNA silent mutation classifier instance."""
        return CodingDNASilentMutationClassifier()

    def fixture_name(self):
        """Return the fixture name for coding DNA silent mutation."""
        return "coding_dna_silent_mutation"



"""A module for testing Silent Mutation tokenization."""
import unittest

from variation.tokenizers.caches import AminoAcidCache
from variation.tokenizers import SilentMutation
from .tokenizer_base import TokenizerBase


class TestSilentMutationTokenizer(TokenizerBase, unittest.TestCase):
    """A class for testing Silent Mutation Tokenization."""

    def tokenizer_instance(self):
        """Return Silent Mutation instance."""
        return SilentMutation(AminoAcidCache())

    def token_type(self):
        """Return Silent Mutation token type."""
        return "SilentMutation"

    def fixture_name(self):
        """Return the fixture name for Silent Mutation."""
        return "silent_mutation"



"""Module for Protein Substitution Translation."""
from variation.translators.translator import Translator
from variation.schemas.classification_response_schema import ClassificationType
from variation.schemas.token_response_schema import ProteinSubstitutionToken, Token


class ProteinSubstitution(Translator):
    """The Protein Substitution Translator class."""

    def can_translate(self, type: ClassificationType) -> bool:
        """Return if classification type is Protein Substitution."""
        return type == ClassificationType.PROTEIN_SUBSTITUTION

    def is_token_instance(self, token: Token) -> bool:
        """Return if the token is an Protein Substitution token instance."""
        return isinstance(token, ProteinSubstitutionToken)



"""Module for testing Coding DNA Insertion Translator."""
import unittest

from variation.classifiers import CodingDNAInsertionClassifier
from variation.translators import CodingDNAInsertion
from variation.validators import CodingDNAInsertion as CDNAD_V
from .translator_base import TranslatorBase


class TestCodingDNAInsertionTranslator(TranslatorBase, unittest.TestCase):
    """A class to test the Coding DNA Insertion Translator."""

    def classifier_instance(self):
        """Return coding DNA insertion instance."""
        return CodingDNAInsertionClassifier()

    def validator_instance(self):
        """Return coding DNA delins instance."""
        return CDNAD_V(*self.params)

    def translator_instance(self):
        """Return coding DNA insertion instance."""
        return CodingDNAInsertion()

    def fixture_name(self):
        """Return the fixture name for coding DNA insertion."""
        return "coding_dna_insertion"



"""Module for testing Polypeptide Truncation Validator."""
import unittest

from variation.validators import PolypeptideTruncation
from variation.classifiers import PolypeptideTruncationClassifier
from .validator_base import ValidatorBase


class TestPolypeptideTruncationValidator(ValidatorBase, unittest.TestCase):
    """A class to test the Polypeptide Truncation Validator."""

    def validator_instance(self):
        """Return Polypeptide Truncation instance."""
        return PolypeptideTruncation(*self.aa_params)

    def classifier_instance(self):
        """Return the Polypeptide Truncation classifier instance."""
        return PolypeptideTruncationClassifier()

    def fixture_name(self):
        """Return the fixture name for Polypeptide Truncation."""
        return "polypeptide_truncation"



"""Module for Classification methods."""
from abc import ABC, abstractmethod
from typing import List, Optional

from variation.schemas.classification_response_schema import Classification,\
    ClassificationType
from variation.schemas.token_response_schema import Token


class Classifier(ABC):
    """The Classifier class."""

    @abstractmethod
    def match(self, tokens: List[Token]) -> Optional[Classification]:
        """Return the classification from a list of token matches."""
        raise NotImplementedError

    @abstractmethod
    def classification_type(self) -> ClassificationType:
        """Return the classification type."""
        raise NotImplementedError



"""Module for testing Polypeptide Truncation Classifier."""
import unittest

from variation.classifiers import PolypeptideTruncationClassifier
from .classifier_base import ClassifierBase


class TestPolypeptideTruncationClassifier(ClassifierBase, unittest.TestCase):
    """A class to test the Polypeptide Truncation Classifier."""

    def classifier_instance(self):
        """Return PolypeptideTruncationClassifier instance."""
        return PolypeptideTruncationClassifier()

    def fixture_name(self):
        """Return PolypeptideTruncationClassifier fixture name."""
        return "polypeptide_truncation"



"""A module for Genomic Deletion Range Tokenization."""
from typing import Dict, Optional, List

from pydantic import NonNegativeFloat

from variation.schemas.token_response_schema import GenomicDeletionRangeToken
from variation.tokenizers.deletion_range_base import DeletionRangeBase


class GenomicDeletionRange(DeletionRangeBase):
    """Class for tokenizing deletion range at the genomic coordinate."""

    def _get_parts(self, parts: List) -> NonNegativeFloat:
        """Set parts for genomic deletion range.

        :param List parts: Parts of input string
        """
        if len(parts) != 4:
            return None

        conditions = (
            parts[0].startswith("g.("),
            parts[1].endswith(")"),
            parts[2].startswith("("),
            parts[3].endswith(")del")
        )

        if all(conditions):
            parts[0] = parts[0][3:]
            parts[1] = parts[1][:-1]
            parts[2] = parts[2][1:]
            parts[3] = parts[3][:-4]

            try:
                parts[0] = int(parts[0])
                parts[1] = int(parts[1])
                parts[2] = int(parts[2])
                parts[3] = int(parts[3])
            except ValueError:
                return None
            else:
                prev_val = None
                for i in range(4):
                    val = parts[i]
                    if val not in ["?", None]:
                        if prev_val is not None:
                            if prev_val > val:
                                return None
                    prev_val = val

                self.parts["start_pos1_del"] = parts[0]
                self.parts["start_pos2_del"] = parts[1]
                self.parts["end_pos1_del"] = parts[2]
                self.parts["end_pos2_del"] = parts[3]
        return None

    def return_token(self, params: Dict) -> Optional[GenomicDeletionRangeToken]:
        """Return Genomic Deletion Range token."""
        conditions = (
            params["start_pos1_del"] is not None,
            params["start_pos2_del"] is not None,
            params["end_pos1_del"] is not None,
            params["end_pos2_del"] is not None
        )
        if all(conditions):
            if params["coordinate_type"] == "g":
                return GenomicDeletionRangeToken(**params)



"""Module for testing Codon Table class."""
import pytest

from variation.data_sources.codon_table import CodonTable
from variation.tokenizers.caches import AminoAcidCache


@pytest.fixture(scope="module")
def test_codon_table():
    """Build codon table test fixture."""
    return CodonTable(AminoAcidCache())


def test_get_codons(test_codon_table):
    """Test that _get_codons method works correctly."""
    ala_codons = ["GCA", "GCC", "GCG", "GCU"]
    assert test_codon_table.get_codons("A") == ala_codons
    assert test_codon_table.get_codons("a") == ala_codons
    assert test_codon_table.get_codons("Ala") == ala_codons
    assert test_codon_table.get_codons("ala") == ala_codons
    assert test_codon_table.get_codons("AlA") == ala_codons

    leu_codons = ["CUA", "CUC", "CUG", "CUU", "UUA", "UUG"]
    assert test_codon_table.get_codons("L") == leu_codons
    assert test_codon_table.get_codons("Leu") == leu_codons

    arg_codons = ["AGA", "AGG", "CGA", "CGC", "CGG", "CGU"]
    assert test_codon_table.get_codons("r") == arg_codons
    assert test_codon_table.get_codons("Arg") == arg_codons


def test_dna_to_rna(test_codon_table):
    """Test that dna_to_rna method works correctly."""
    resp = test_codon_table.dna_to_rna("GTA")
    assert resp == "CAU"

    resp = test_codon_table.dna_to_rna("AAGTGACA")
    assert resp == "UUCACUGU"



"""A module for the Coding DNA Deletion Classifier."""
from typing import List

from variation.schemas.classification_response_schema import ClassificationType
from .set_based_classifier import SetBasedClassifier


class CodingDNADeletionClassifier(SetBasedClassifier):
    """The Coding DNA Deletion Classifier class."""

    def classification_type(self) -> ClassificationType:
        """Return the Coding DNA Deletion classification type."""
        return ClassificationType.CODING_DNA_DELETION

    def exact_match_candidates(self) -> List[List[str]]:
        """Return the exact match token type candidates."""
        return [
            ["CodingDNADeletion"],
            ["GeneSymbol", "ProteinSubstitution", "CodingDNADeletion"],
            ["CodingDNADeletion", "GeneSymbol"],
            ["GeneSymbol", "CodingDNADeletion"],
            ["HGVS", "CodingDNADeletion"],
            ["ReferenceSequence", "CodingDNADeletion"],
            ["LocusReferenceGenomic", "CodingDNADeletion"]
        ]



"""A module for testing Genomic Insertion Tokenization."""
import unittest

from variation.tokenizers.caches import AminoAcidCache, NucleotideCache
from variation.tokenizers import GenomicInsertion
from .tokenizer_base import TokenizerBase


class TestGenomicInsertionTokenizer(TokenizerBase, unittest.TestCase):
    """A class for testing Genomic Insertion Tokenization."""

    def tokenizer_instance(self):
        """Return Genomic Insertion instance."""
        return GenomicInsertion(AminoAcidCache(), NucleotideCache())

    def token_type(self):
        """Return genomic insertion token type."""
        return "GenomicInsertion"

    def fixture_name(self):
        """Return the fixture name for Genomic Insertion."""
        return "genomic_insertion"



"""Module for testing Coding DNA DelIns Translator."""
import unittest

from variation.classifiers import CodingDNADelInsClassifier
from variation.translators import CodingDNADelIns
from variation.validators import CodingDNADelIns as CDNADELINS_V
from .translator_base import TranslatorBase


class TestCodingDNADelInsTranslator(TranslatorBase, unittest.TestCase):
    """A class to test the Coding DNA DelIns Translator."""

    def classifier_instance(self):
        """Return coding DNA delins instance."""
        return CodingDNADelInsClassifier()

    def validator_instance(self):
        """Return coding DNA delins instance."""
        return CDNADELINS_V(*self.params)

    def translator_instance(self):
        """Return coding DNA delins instance."""
        return CodingDNADelIns()

    def fixture_name(self):
        """Return the fixture name for coding DNA delins."""
        return "coding_dna_delins"
